from .interfaceMathsToPhysics import InterfaceMathsToPhysics
# from optimeed.core import printIfShown, SHOW_WARNING


class MathsToPhysics(InterfaceMathsToPhysics):
    """Dummy yet powerful example of maths to physics.
    The optimization variables are directly injected to the device"""
    def __init__(self):
        super().__init__()

    def fromMathsToPhys(self, xVector, theDevice, theOptimizationVariables):
        if len(xVector) != len(theOptimizationVariables):
            raise ValueError("In mathsToPhysics: length vectors and optimization variables are not the same")

        for i in range(len(xVector)):
            theOptimizationVariables[i].do_MathsToPhys(xVector[i], theDevice)

    def fromPhysToMaths(self, theDevice, theOptimizationVariables):
        x01 = [None]*len(theOptimizationVariables)
        for i, optimizationVariable in enumerate(theOptimizationVariables):
            x01[i] = optimizationVariable.get_PhysToMaths(theDevice)
        return x01

    def __str__(self):
        return "Dummy maths to physics"



from optimeed.core import printIfShown, SHOW_WARNING
import traceback


def evaluate(inputs):
    x, theDevice, theMathsToPhys, theCharacterization, list_of_optimization_variables, index = inputs[0], inputs[1], inputs[2], inputs[3], inputs[4], inputs[5]
    """Main evaluation function"""
    output = dict()
    output["device"] = theDevice
    output["x"] = x
    output["index"] = index

    theMathsToPhys.fromMathsToPhys(x, theDevice, list_of_optimization_variables)
    # noinspection PyBroadException
    try:
        theCharacterization.compute(theDevice)
        output["success"] = True
    except Exception:
        printIfShown("An error in characterization. Bypassing it to continue execution. Error :" + traceback.format_exc(), SHOW_WARNING)
        output["success"] = False
    return output



from PyQt5 import QtWidgets, Qt
from PyQt5.QtCore import pyqtSlot, pyqtSignal
from PyQt5.Qt import QAbstractItemView


class Widget_tableWithSearch(QtWidgets.QWidget):
    cellChanged = pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        layout_widget = QtWidgets.QVBoxLayout(self)
        layout_widget.setContentsMargins(0, 0, 0, 0)

        self.myTableWidget = QtWidgets.QTableWidget()

        # self.myTableWidget.setSelectionMode(QAbstractItemView.SingleSelection)
        self.myTableWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.search_plot_data = QtWidgets.QLineEdit()
        # self.search_plot_data.setTabChangesFocus(True)

        self.layout().addWidget(self.search_plot_data)
        self.layout().addWidget(self.myTableWidget)
        self.forced_hidden_rows = list()

        # Connect signals
        self.myTableWidget.cellChanged.connect(self._cellChanged)
        self.search_plot_data.textChanged.connect(self._filter_list)

    def hideRow(self, row):
        self.myTableWidget.hideRow(row)

    def showRow(self, row):
        self.myTableWidget.showRow(row)

    def force_hide_row(self, row):
        self.forced_hidden_rows.append(row)
        self.hideRow(row)

    def remove_forced_hide_row(self, row):
        self.showRow(row)
        try:
            self.forced_hidden_rows.remove(row)
        except ValueError:
            pass

    @pyqtSlot()
    def get_entries_selected(self):
        results = []
        rows = []
        for index in sorted(self.myTableWidget.selectionModel().selectedRows()):
            row = index.row()
            rows.append(row)
            results_temp = [self.myTableWidget.item(row, k) for k in range(self.myTableWidget.columnCount())]
            results_temp = [res if res is None else res.data(0) for res in results_temp]
            results.append(results_temp)
        return results, rows

    def _cellChanged(self):
        self.cellChanged.emit()

    def set_entries(self, names, numColumns=3, hidden=False):
        self.myTableWidget.setRowCount(len(names))
        self.myTableWidget.setColumnCount(numColumns)
        for k, name in enumerate(names):
            item = QtWidgets.QTableWidgetItem(name)
            item.setFlags(item.flags() ^ Qt.Qt.ItemIsEditable)
            self.myTableWidget.setItem(k, 0, item)
            if hidden:
                self.forced_hidden_rows.append(k)
                self.hideRow(k)

        header = self.myTableWidget.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)

    def get_shown_entries(self):
        results = []
        rows = []
        for row in range(self.myTableWidget.model().rowCount()):
            if row not in self.forced_hidden_rows:
                rows.append(row)
                results_temp = [self.myTableWidget.item(row, k) for k in range(self.myTableWidget.columnCount())]
                results_temp = [res if res is None else res.data(0) for res in results_temp]
                results.append(results_temp)
        return results, rows

    def set_item(self, row, col, item):
        self.myTableWidget.setItem(row, col, QtWidgets.QTableWidgetItem(item))

    def get_item(self, row, col):
        return self.myTableWidget.item(row, col)

    def _filter_list(self):
        text = self.search_plot_data.text()
        for k, item in self._iter_items():
            if text in item.text() and k not in self.forced_hidden_rows:
                self.myTableWidget.showRow(k)
            else:
                self.myTableWidget.hideRow(k)

    def _iter_items(self):
        for i in range(self.myTableWidget.rowCount()):
            yield i, self.myTableWidget.item(i, 0)



from abc import ABCMeta, abstractmethod


class InterfaceObjCons(metaclass=ABCMeta):
    """Interface class for objectives and constraints. The objective is to MINIMIZE and the constraint has to respect VALUE <= 0"""

    @abstractmethod
    def compute(self, theDevice):
        """
        Get the value of the objective or the constraint. The objective is to MINIMIZE and the constraint has to respect VALUE <= 0

        :param theDevice: Input device that has already been evaluated
        :return: float.
        """
        pass

    def get_name(self):
        return "Objective (or constraint)"

    def __str__(self):
        return self.get_name()



import os
import traceback
from PyQt5 import QtCore, QtWidgets

from optimeed.core.inkscape_manager import inkscape_svg_to_pdf, inkscape_svg_to_png
from optimeed.core.tools import printIfShown, SHOW_WARNING
from optimeed.core.tikzTranslator import export_to_tikz_groupGraphs

from .pyqtgraph import setConfigOption, setConfigOptions
from .pyqtgraph.exporters.SVGExporter import SVGExporter
from .pyqtgraphRedefine import myGraphicsLayoutWidget

from .traceVisual import TraceVisual
from .graphVisual import GraphVisual

from optimeed.visualize.widgets import Widget_menuButton


class Widget_graphsVisualLite(QtWidgets.QWidget):
    """Widget element to draw a graph. The traces and graphs to draw are defined in :class:`~optimeed.visualize.graphs.Graphs.Graphs` taken as argument.
     This widget is linked to the excellent third-party library pyqtgraph, under MIT license"""
    signal_must_update = QtCore.pyqtSignal()  # Use this signal to refresh the graphs (signal_must_update.emit())
    signal_graph_changed = QtCore.pyqtSignal()  # This signal is sent when the graph layout is changed (number of traces/graphs)

    def __init__(self, theGraphs, **kwargs):
        """

        :param theGraphs: :class:`~optimeed.visualize.graphs.Graphs.Graphs`.
        :param is_light: Light theme option (bool)
        :param refresh_time: Refresh time (in seconds). If < 0: never automatically refreshed.
        :param highlight_last: Option to automatically highlight last point added in data.
        :param linkX: Link the x-axis of the graphs together (bool)
        :param actionOnClick: Action to perform when a point is clicked. Calls the method :meth:`~on_graph_click_interface.graph_clicked`.
        """
        self.is_light = kwargs.get('is_light', False)
        self.refresh_time = kwargs.get('refresh_time', 0.1) * 1000
        self.highlight_last = kwargs.get('highlight_last', False)
        self.linkX = kwargs.get('linkX', False)
        self.theActionOnClick = kwargs.get('actionOnClick', None)

        super().__init__()

        # Configure pyqtgraphs
        if self.is_light:  # Template "Light"
            background_color = 'w'
            foreground_color = (50, 50, 50)
            self.label_color = '#000'
        else:
            background_color = (30, 30, 30)
            foreground_color = (200, 200, 200)
            self.label_color = '#FFF'

        setConfigOption("antialias", True)
        setConfigOption('background', background_color)  # Color of background
        setConfigOption('foreground', foreground_color)  # Color of ticks
        setConfigOptions(useOpenGL=True)  # Otherwise performance issue on line thickness
        # pg.setConfigOption("crashWarning", True)

        # Initializes upper widget and core variables

        main_vertical_layout = QtWidgets.QVBoxLayout(self)
        self.canvasWidget = myGraphicsLayoutWidget()
        main_vertical_layout.addWidget(self.canvasWidget)


        # self.theTraces = dict()
        # self.theWGPlots = dict()
        self.theGraphs = dict()
        self.theGraphsVisual = dict()

        # Create graphs
        self.theGraphs = theGraphs
        theGraphs.add_update_method(lambda: self.signal_must_update.emit())

        if self.linkX:
            self.link_axes()

        self.update_graphs(singleUpdate=False)
        self.canvasWidget.showMaximized()

        # connect signals
        self.signal_must_update.connect(self.update_graphs)

    def set_graph_disposition(self, indexGraph, row=1, col=1, rowspan=1, colspan=1):
        """
        Change the graphs disposition.

        :param indexGraph: index of the graph to change
        :param row: row where to place the graph
        :param col: column where to place the graph
        :param rowspan: number of rows across which the graph spans
        :param colspan: number of columns across which the graph spans
        :return:
        """
        item = self.get_graph(indexGraph).theWGPlot
        self.canvasWidget.ci.set_graph_disposition(item, row, col, rowspan, colspan)

    def __create_graph(self, idGraph):
        self.theGraphsVisual[idGraph] = GraphVisual(self)

    def __check_graphs(self):
        graph_changed = False
        """If the graphs have been modified, update them."""
        graphs2D = self.theGraphs.get_all_graphs_ids()
        # # Check the graphs

        # Create non existing graphs
        for idGraph in graphs2D:
            if idGraph not in self.theGraphsVisual:
                graph_changed = True
                self.__create_graph(idGraph)  # Create a graph
        # Remove graphs that disappeared
        indexToRemove = []
        for idGraph in self.theGraphsVisual:
            if idGraph not in graphs2D:
                graph_changed = True
                indexToRemove.append(idGraph)
        for idGraph in indexToRemove:
            self.delete_graph(idGraph)

        # # Check the traces
        for idGraph in self.theGraphsVisual:
            graphVisual = self.theGraphsVisual[idGraph]
            traces = self.theGraphs.get_graph(idGraph).get_all_traces()
            tracesVisual = graphVisual.get_all_traces()
            # Create non existing traces
            for idTrace, trace in traces.items():
                if idTrace not in tracesVisual:
                    graph_changed = True
                    graphVisual.add_data(idTrace, trace)  # Create the traces

            # Remove traces that  disappeared
            indexToRemove = []
            for idTrace in tracesVisual:
                if idTrace not in traces:
                    graph_changed = True
                    indexToRemove.append(idTrace)

            for idTrace in indexToRemove:
                graphVisual.delete_trace(idTrace)

            # Set color
            graphVisual.apply_palette()

            graphVisual.set_legend()

            if graph_changed:
                self.signal_graph_changed.emit()

    def on_click(self, plotDataItem, clicked_points):  # Clicked_points = list of spotItem
        indices_points = [0]*len(clicked_points)
        index_graph = 0
        index_trace = 0

        # Get index of point

        all_points = plotDataItem.scatter.points().tolist()
        for i, point in enumerate(clicked_points):
            indices_points[i] = all_points.index(point)

        # get index of graph
        for idGraph, graphVisual in self.theGraphsVisual.items():
            for idTrace, traceVisual in graphVisual.get_all_traces().items():
                if traceVisual.thePlotItem == plotDataItem:
                    index_graph = idGraph
                    index_trace = idTrace

        if self.theActionOnClick is not None:
            # noinspection PyBroadException
            try:
                self.theActionOnClick.graph_clicked(self, index_graph, index_trace, indices_points)
            except Exception:
                printIfShown("Following error occurred in action on click :" + traceback.format_exc(), SHOW_WARNING)

    def update_graphs(self, singleUpdate=True):
        """
        This method is used to update the graph. This is fast but NOT safe (especially when working with threads).
        To limit the risks, please use self.signal_must_update.emit() instead.

        :param singleUpdate: if set to False, the graph will periodically refres each self.refreshtime
        """
        self.__check_graphs()
        # self.fast_update()
        for graph in self.theGraphsVisual.values():
            graph.update()
        if self.refresh_time > 0 and not singleUpdate:
            QtCore.QTimer().singleShot(self.refresh_time, lambda: self.update_graphs(False))

    def fast_update(self):
        """Use this method to update the graph in a fast way. NOT THREAD SAFE."""
        for graph in self.theGraphsVisual.values():
            graph.fast_update()

    def select_folder_and_export(self):
        dlg = QtWidgets.QFileDialog.getSaveFileName()[0]
        self.exportGraphs(dlg)

    def exportGraphs(self, filename):
        """Export the graphs"""
        if filename:
            root, ext = os.path.splitext(filename)
            ext_svg = '.svg'
            ext_png = '.png'
            ext_pdf = '.pdf'
            ext_txt = '.txt'
            ext_tikz = '.tikz'

            if ext == ext_txt or not ext:
                self.export_txt(root + ext_txt)

            if ext == ext_svg or not ext or ext in [ext_png, ext_pdf]:
                self.export_svg(root + ext_svg)

            if ext == ext_png or not ext:
                inkscape_svg_to_png(root + ext_svg, root + ext_png)

            if ext == ext_pdf or not ext:
                inkscape_svg_to_pdf(root + ext_svg, root + ext_pdf)

            if ext == ext_tikz or not ext:
                self.export_tikz(root)

    def export_txt(self, filename_txt):
        theStr = self.theGraphs.export_str()
        with open(filename_txt, 'w') as f:
            f.write(theStr)

    def export_svg(self, filename):
        my_exporter = SVGExporter(self.canvasWidget.sceneObj)
        my_exporter.params.param('scaling stroke').setValue(True)
        my_exporter.export(filename)

    def export_tikz(self, foldername_tikz):
        # Additional parameters for tikz: log axes
        def additionalAxisOptions(graphId):
            theStr = ''
            gridstyle = "grid style={solid},"
            if self.get_graph(graphId).theWGPlot.ctrl.logXCheck.isChecked():
                theStr += "xmode=log,xtickten={0,...,10}," + gridstyle
            if self.get_graph(graphId).theWGPlot.ctrl.logYCheck.isChecked():
                theStr += "ymode=log,ytickten={0,...,10}," + gridstyle
            # axisX = self.get_graph(graphId).get_axis("bottom")
            # axisY = self.get_graph(graphId).get_axis("left")
            # xrange = axisX.range
            # yrange = axisY.range
            # theStr += "xmin={}, xmax={}\n, ymin={}, ymax={}\n".format(*xrange, *yrange)
            return theStr

        # def additionalTraceOptions(graphId, traceId):
        #     theStr = ''
        #     # theColor = self.get_graph(graphId).get_trace(traceId).get_color()
        #     # if self.theGraphs.get_graph(graphId).get_trace(traceId).get_color() is None:
        #     #     theStr += "color={{rgb:red,{};green,{};blue,{}}},\n".format(*convert_color(theColor))
        #     return theStr
        export_to_tikz_groupGraphs(self.theGraphs, foldername_tikz, additionalAxisOptions=additionalAxisOptions)

    def link_axes(self):
        keys = list(self.theGraphsVisual.keys())
        for key in keys:
            self.theGraphsVisual[key].linkXToGraph(self.theGraphsVisual[keys[0]])

    def get_graph(self, idGraph) -> GraphVisual:
        """Get corresponding :class:`~optimeed.visualize.gui.widgets.graphsVisualWidget.GraphVisual.GraphVisual` of the graph idGraph"""
        return self.theGraphsVisual[idGraph]

    def get_trace(self, idGraph, idTrace) -> TraceVisual:
        """Get corresponding Tracevisual"""
        return self.get_graph(idGraph).get_trace(idTrace)

    def keyPressEvent(self, event):
        """
        What happens if a key is pressed.
        R: reset the axes to their default value
        """
        if event.key() == QtCore.Qt.Key_R:
            for idGraphVisual in self.get_all_graphsVisual():
                self.get_graph(idGraphVisual).theWGPlot.autoRange()
                # graphVisual.set_lims(self.theTraces[idGraph][0])

    def delete_graph(self, idGraph):
        """Delete the graph idGraph"""
        try:
            self.theGraphsVisual[idGraph].delete()
            self.theGraphsVisual.pop(idGraph)
        except KeyError:
            printIfShown("Graph already deleted", SHOW_WARNING)

    def delete(self):
        graphs_ids = list(self.get_all_graphsVisual().keys())
        for graphID in graphs_ids:
            self.delete_graph(graphID)
        self.deleteLater()

    def get_all_graphsVisual(self):
        """Return a dictionary {idGraph: :class:`~optimeed.visualize.gui.widgets.graphsVisualWidget.GraphVisual.GraphVisual`}."""
        return self.theGraphsVisual

    def get_layout_buttons(self):
        """Get the QGraphicsLayout where it's possible to add buttons, etc."""
        return self.horizontalLayoutGUI

    def set_actionOnClick(self, theActionOnClick):
        """Action to perform when the graph is clicked

        :param theActionOnClick: :class:`on_graph_click_interface`
        :return:
        """
        self.theActionOnClick = theActionOnClick

    def set_title(self, idGraph, titleName, **kwargs):
        """Set title of the graph

        :param idGraph: id of the graph
        :param titleName: title to set
        """
        self.get_graph(idGraph).set_title(titleName, **kwargs)


class Widget_graphsVisual(Widget_graphsVisualLite):
    """Create a gui for pyqtgraph with trace selection options, export and action on clic choices"""
    def __init__(self, *args, **kwargs):
        """See Widget_graphsVisualLite params. Extra:

        :param actionsOnClick: list of  :class:`~optimeed.visualize.onclick.onclickInterface.OnclickInterface`
        """
        super().__init__(*args, **kwargs)


        # # Add last buttons
        # Export button
        horizontalLayout = QtWidgets.QHBoxLayout()

        # self.centralLayout
        self.layout().addLayout(horizontalLayout)
        # main_vertical_layout.addLayout(horizontalLayoutGUI)

        # self.theGraphsVisual = graphsVisual
        self.actionsOnClick = kwargs.pop('actionsOnClick', list())

        # Define buttons
        # horizontalLayout = self.theGraphsVisual.get_layout_buttons()

        button = QtWidgets.QPushButton('Export Graphs')
        button.clicked.connect(self.select_folder_and_export)
        horizontalLayout.addWidget(button)

        # Action button
        self.actionSelector = QtWidgets.QComboBox()
        self.actionSelector.currentIndexChanged.connect(lambda index: self.set_actionOnClick(self.actionsOnClick[index]))

        if self.actionsOnClick:
            self.set_actions_on_click(self.actionsOnClick)

        horizontalLayout.addWidget(self.actionSelector)

        # Link axes button
        button = QtWidgets.QPushButton('Link axes')
        button.clicked.connect(self.link_axes)
        horizontalLayout.addWidget(button)

        # Traces manager button
        button = QtWidgets.QPushButton('Manage traces')
        self.traceManager = Widget_menuButton(button)
        horizontalLayout.addWidget(button)
        self.signal_graph_changed.connect(lambda: self.refreshTraceList())
        self.refreshTraceList()

    def refreshTraceList(self):
        """Refresh all the traces"""
        self.traceManager.clear()
        graphsVisuals = self.get_all_graphsVisual()
        for graphId in self.get_all_graphsVisual():
            theGraph = graphsVisuals[graphId]
            all_traces = theGraph.get_all_traces()
            for trace in all_traces.values():
                theAction = self.traceManager.addAction("[{}] ".format(graphId) + trace.get_data().get_legend())
                theAction.setCheckable(True)
                theAction.setChecked(True)
                theAction.toggled.connect(trace.toggle)

    def set_actions_on_click(self, actions):
        self.actionsOnClick = actions
        for action in self.actionsOnClick:
            self.actionSelector.addItem(action.get_name())
        self.actionSelector.setCurrentIndex(0)
        self.set_actionOnClick(self.actionsOnClick[0])



from .pyswarm import MyMapEvaluator, MyMultiprocessEvaluator, pso

from optimeed.core.tools import indentParagraph  # isNonePrintMessage, printIfShown, SHOW_ERROR, SHOW_WARNING
from .algorithmInterface import AlgorithmInterface
from optimeed.core import Option_class, Option_int
from optimeed.optimize.optiAlgorithms.convergence.evolutionaryConvergence import EvolutionaryConvergence


class Monobjective_PSO(AlgorithmInterface, Option_class):
    NUMBER_OF_CORES = 1

    def __init__(self):
        super().__init__()
        self.maxTime = None  # set by set_maxtime
        self.callback_on_evaluation = None  # set by set_evaluationFunction
        self.evaluationFunction = None  # set by set_evaluationFunction
        self.add_option(self.NUMBER_OF_CORES, Option_int("Number of cores used in evaluation", 1))
        self.theConvergence = EvolutionaryConvergence()

        self.initialVectorGuess = None
        self.listOfOptimizationVariables = None

    def initialize(self, initialVectorGuess, listOfOptimizationVariables):
        self.initialVectorGuess = initialVectorGuess
        self.listOfOptimizationVariables = listOfOptimizationVariables

    def compute(self):
        nbr_cores = self.get_option_value(self.NUMBER_OF_CORES)
        # Set the evaluator
        if nbr_cores == 1:
            theEvaluator = MyMapEvaluator(self.evaluationFunction, self.callback_on_evaluation)
        else:
            theEvaluator = MyMultiprocessEvaluator(self.evaluationFunction, self.callback_on_evaluation, nbr_cores)

        # Get lower bounds and upper bounds
        lb = [optiVariable.get_min_value() for optiVariable in self.listOfOptimizationVariables]
        ub = [optiVariable.get_max_value() for optiVariable in self.listOfOptimizationVariables]

        # Run the optimization algorithm
        kwargs = dict()
        optimal_parameters, _function_value, is_feasible = pso(lb, ub, self.initialVectorGuess, theEvaluator, self.maxTime,
                                                               callback_generation=self.theConvergence.set_curr_step,
                                                               **kwargs)
        return [optimal_parameters]

    def set_evaluationFunction(self, evaluationFunction, callback_on_evaluate, numberOfObjectives, _numberOfConstraints):
        if numberOfObjectives > 1:
            ValueError("Optimization algorithm does not support true multiobjective.")

        self.evaluationFunction = evaluationFunction
        self.callback_on_evaluation = callback_on_evaluate

    def set_maxtime(self, maxTime):
        self.maxTime = maxTime

    def __str__(self):
        theStr = ''
        theStr += "Custom PSO\n"
        theStr += indentParagraph(super().__str__(), indent_level=1)
        return theStr

    def get_convergence(self):
        return self.theConvergence



from math import *
import numpy


def normalize(v, tolerance=0.001):
    mag2 = sum(n * n for n in v)
    if abs(mag2 - 1.0) > tolerance:
        mag = sqrt(mag2)
        v = tuple(n / mag for n in v)
    return v


def q_mult(q1, q2):
    w1, x1, y1, z1 = q1
    w2, x2, y2, z2 = q2
    w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
    x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
    y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
    z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
    return w, x, y, z


def q_conjugate(q):
    w, x, y, z = q
    return w, -x, -y, -z


def qv_mult(q1, v1):
    q2 = (0.0,) + v1
    return q_mult(q_mult(q1, q2), q_conjugate(q1))[1:]


def axisangle_to_q(v, theta):
    v = normalize(v)
    x, y, z = v
    theta /= 2
    w = cos(theta)
    x = x * sin(theta)
    y = y * sin(theta)
    z = z * sin(theta)
    return w, x, y, z


def q_to_axisangle(q):
    w, v = q[0], q[1:]
    theta = acos(w) * 2.0
    return normalize(v), theta


def q_to_mat4(q):
    w, x, y, z = q
    return numpy.array([[1 - 2*y*y - 2*z*z, 2*x*y - 2*z*w, 2*x*z + 2*y*w, 0], [2*x*y + 2*z*w, 1 - 2*x*x - 2*z*z, 2*y*z - 2*x*w, 0], [2*x*z - 2*y*w, 2*y*z + 2*x*w, 1 - 2*x*x - 2*y*y, 0], [0, 0, 0, 1]], 'f')



from optimeed.core import Graphs, Data, ListDataStruct_Interface
import numpy as np
from optimeed.core import SHOW_INFO, printIfShown, SHOW_ERROR


class HowToPlotGraph:
    def __init__(self, attribute_x, attribute_y, kwargs_graph=None, check_if_plot_elem=None, meta=None):
        """
        Tells to LinkDataGraph how to plot it

        :param attribute_x: attribute (as string) to plot in X axis
        :param attribute_y: attribute (as string) to plot in Y axis
        :param kwargs_graph: kwargs (as dictionary) of plot options
        :param check_if_plot_elem: function taking an item of the collection as input and outputting True if item has to be displayed.
        :param meta:
        """
        self.attribute_x = attribute_x
        self.attribute_y = attribute_y
        self.meta = meta

        if kwargs_graph is None:
            self.kwargs_graph = dict()
        else:
            self.kwargs_graph = kwargs_graph

        self.check_if_plot_elem = check_if_plot_elem

    def __str__(self):
        theStr = ''
        theStr += "x: {} \t y: {}".format(self.attribute_x, self.attribute_y)
        return theStr


class LinkDataGraph:
    def __init__(self):
        self.theGraphs = Graphs()
        self.collections = dict()  # keys are unique ids of collection
        self.shadow_collections = dict()  # Keys are unique ids of collection
        self.kwargs_collections = dict()  # keys are unique ids of collection
        self.howToPlotGraphs = dict()  # keys are ids of graph
        self.ids_graphs_to_col = dict()  # Dict of dicts. First keys are ids of graphs, second keys are ids of traces.
        self.curr_id = 0

    # Main user function to use the class
    def add_collection(self, theCollection, kwargs=None):
        """ Add a collection (that will be a future trace)

        :param theCollection:
        :param kwargs: kwargs associated with the collection (e.g., color, symbol style, etc.)
        :return: unique id associated with the collection
        """
        theId = self.curr_id
        if kwargs is None:
            kwargs = dict()
        self.kwargs_collections[theId] = kwargs
        self.collections[theId] = theCollection
        self.curr_id += 1
        self.update_graphs()
        return theId

    def remove_collection(self, collectionId):
        """
        Remove collection from the graphs

        :param collectionId: ID of the collection
        :return:
        """
        # Remove from trace
        res, _ = self.get_graph_and_trace_from_idCollection(collectionId)
        for idGraph, idTrace in res:
            self.theGraphs.get_graph(idGraph).remove_trace(idTrace)
            self.ids_graphs_to_col[idGraph].pop(idTrace, None)
        # Remove collection
        self.collections.pop(collectionId, None)
        self.kwargs_collections.pop(collectionId, None)
        self.shadow_collections.pop(collectionId, None)
        self.update_graphs()

    def set_shadow_collection(self, master_collectionId, shadow_collection):
        """
        Link a collection to an other

        :param master_collectionId: ID of the collection that is displayed in the graph
        :param shadow_collection: collection to link to the master.
        :return:
        """
        self.shadow_collections[master_collectionId] = shadow_collection

    def get_graphs(self):
        return self.theGraphs

    def get_howToPlotGraph(self, idGraph):
        return self.howToPlotGraphs[idGraph]

    def add_graph(self, howToPlotGraph):
        """Add new graph to be plotted.

        :param howToPlotGraph: :class:`HowToPlotGraph`
        :return:
        """
        idGraph = self.theGraphs.add_graph()
        self.howToPlotGraphs[idGraph] = howToPlotGraph
        self.ids_graphs_to_col[idGraph] = dict()
        self.update_graphs()
        return idGraph

    def get_idCollections(self):
        """Get all ids of the plotted collections"""
        return list(self.collections.keys())

    def get_idGraphs(self):
        """Get all ids of the graphs"""
        return list(self.ids_graphs_to_col.keys())

    def get_idTraces(self, idGraph):
        """Get all ids of the traces of graph $idGraph"""
        return list(self.ids_graphs_to_col[idGraph].keys())

    def get_idCollection_from_graph(self, idGraph, idTrace):
        """Get id of collection plotted in graph $idGraph and trace $idTrace"""
        return self.ids_graphs_to_col[idGraph][idTrace]

    def get_collection(self, idCollection, getShadow=True):
        if getShadow:
            try:
                return self.shadow_collections[idCollection]
            except KeyError:
                pass
        return self.collections[idCollection]

    @staticmethod
    def get_x_y_to_plot(theCollection, howToPlotGraph):
        """Extract X-Y infos from collection and howtoplotGraph"""
        y_data = theCollection.get_list_attributes(howToPlotGraph.attribute_y)
        if howToPlotGraph.attribute_x is None:
            x_data = list(range(len(y_data)))
        else:
            x_data = theCollection.get_list_attributes(howToPlotGraph.attribute_x)
        min_length = min(len(x_data), len(y_data))
        return x_data[:min_length], y_data[:min_length]  # Truncate if lengths are not the same

    @staticmethod
    def get_meta_to_plot(theCollection, howToPlotGraph):
        """Get Z info from collection and HowToPlotGraph"""
        if howToPlotGraph.meta is None:
            return None
        else:
            return theCollection.get_list_attributes(howToPlotGraph.meta)

    def update_graphs(self):
        """Update the graphs: update graphs, traces, and X-Y data"""
        for idGraph in self.get_idGraphs():
            howToPlotGraph = self.howToPlotGraphs[idGraph]

            for idCollection in self.collections:
                # Check if collection belongs in traces, if not create it
                if idCollection not in self.ids_graphs_to_col[idGraph].values():
                    theData = Data([], [])
                    idTrace = self.theGraphs.add_trace(idGraph, theData, updateChildren=False)
                    self.ids_graphs_to_col[idGraph][idTrace] = idCollection

            for idTrace in self.get_idTraces(idGraph):
                idCollection = self.get_idCollection_from_graph(idGraph, idTrace)
                theData = self.theGraphs.get_graph(idGraph).get_trace(idTrace)
                theCollection = self.get_collection(idCollection, getShadow=False)
                try:
                    x, y = self.get_x_y_to_plot(theCollection, howToPlotGraph)
                except AttributeError as e:
                    x, y = [], []
                    printIfShown("{}".format(e), SHOW_ERROR)

                # Replace infinite and nan numbers by 0
                vec = [x, y]
                for i in range(2):
                    # Replace "inf" by "nan
                    vec[i] = np.array(vec[i], dtype=float)
                    vec[i][np.isinf(vec[i])] = float('nan')
                    pos_infinite = np.argwhere(np.logical_not(np.isfinite(vec[i])))
                    vec[i][pos_infinite] = 0
                    if all(np.isnan(vec[i])):
                        vec[i] = []
                    else:
                        vec[i] = vec[i].tolist()
                x, y = vec[0], vec[1]

                kwargs = dict()
                kwargs.update(howToPlotGraph.kwargs_graph)
                kwargs.update(self.kwargs_collections[idCollection])

                theData.set_data(x, y)
                theData.set_meta(self.get_meta_to_plot(theCollection, howToPlotGraph))
                theData.set_kwargs(kwargs)

                indices_to_plot = list()
                if howToPlotGraph.check_if_plot_elem is not None:
                    for k in range(len(theCollection)):
                        if howToPlotGraph.check_if_plot_elem(theCollection.get_data_at_index(k)) and k < len(x):
                            indices_to_plot.append(k)
                    theData.set_indices_points_to_plot(indices_to_plot)
        self.theGraphs.updateChildren()

    # Direct search functions: from graphs -> to collection

    def get_collection_from_graph(self, idGraph, idTrace, getShadow=True) -> ListDataStruct_Interface:
        """From indices in the graph, get corresponding collection"""
        return self.get_collection(self.get_idCollection_from_graph(idGraph, idTrace), getShadow=getShadow)

    def get_clicked_item(self, idGraph, idTrace, idPoint, getShadow=True):
        """
        Get the data hidden behind the clicked point

        :param idGraph: ID of the graph
        :param idTrace: ID of the trace
        :param idPoint: ID of the point
        :param getShadow: If true, will return the data from the collection linked to the collection that is plotted
        :return: Object in collection
        """
        theCol = self.get_collection_from_graph(idGraph, idTrace, getShadow)
        theData = self.theGraphs.get_graph(idGraph).get_trace(idTrace)
        index_in_data = theData.get_dataIndex_from_graphIndex(idPoint)
        return theCol.get_data_at_index(index_in_data)

    def get_clicked_items(self, idGraph, idTrace, idPoint_list, getShadow=True):
        """Same as get_clicked_item, but using a list of points"""
        theCol = self.get_collection_from_graph(idGraph, idTrace, getShadow)
        theData = self.theGraphs.get_graph(idGraph).get_trace(idTrace)
        indices_in_data = theData.get_dataIndices_from_graphIndices(idPoint_list)
        return [theCol.get_data_at_index(index_in_data) for index_in_data in indices_in_data]

    def delete_clicked_item(self, idGraph, idTrace, idPoint):
        """Remove item from the collection"""
        theData = self.theGraphs.get_graph(idGraph).get_trace(idTrace)
        index_in_data = theData.get_dataIndex_from_graphIndex(idPoint)
        getShadows = [True, False] if self.get_idCollection_from_graph(idGraph, idTrace) in self.shadow_collections else [False]
        for getShadow in getShadows:
            self.get_collection_from_graph(idGraph, idTrace, getShadow).delete_points_at_indices([index_in_data])
        self.update_graphs()

    def delete_clicked_items(self, idGraph, idTrace, idPoints):
        """Same, but for a list of points"""
        theData = self.theGraphs.get_graph(idGraph).get_trace(idTrace)
        indices_in_data = theData.get_dataIndices_from_graphIndices(idPoints)
        getShadows = [True, False] if self.get_idCollection_from_graph(idGraph, idTrace) in self.shadow_collections else [False]
        for getShadow in getShadows:
            self.get_collection_from_graph(idGraph, idTrace, getShadow).delete_points_at_indices(indices_in_data)
        self.update_graphs()

    # Functions here are reverse search: from collection -> find graphs
    def get_graph_and_trace_from_idCollection(self, idCollection):
        """Reverse search: from a collection, get all associated graphs"""
        res = list()
        for idGraph in self.get_idGraphs():
            for idTrace in self.get_idTraces(idGraph):
                if self.get_idCollection_from_graph(idGraph, idTrace) == idCollection:
                    res.append((idGraph, idTrace))
        return res, self.collections[idCollection]

    def get_idcollection_from_collection(self, theCollection):
        """Reverse search: from a collection, find its id"""
        for idCollection in self.collections:
            if self.collections[idCollection] == theCollection:
                return idCollection

    def get_idPoints_from_indices_in_collection(self, idGraph, idTrace, indices_in_collection):
        """From indices in a collection, find the associated idPoints of the graph"""
        theData = self.theGraphs.get_graph(idGraph).get_trace(idTrace)
        return theData.get_graphIndices_from_dataIndices(indices_in_collection)



# Copyright 2015-2018 David Hadka
#
# This file is part of Platypus, a Python module for designing and using
# evolutionary algorithms (EAs) and multiobjective evolutionary algorithms
# (MOEAs).
#
# Platypus is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Platypus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Platypus.  If not, see <http://www.gnu.org/licenses/>.
import copy
import random
import unittest
from ..core import Constraint, Problem, Solution, ParetoDominance, Archive, \
        nondominated_sort, nondominated_truncate, nondominated_prune, \
        POSITIVE_INFINITY, nondominated_split, truncate_fitness, normalize, \
        EpsilonBoxArchive

def createSolution(*args):
    problem = Problem(0, len(args))
    solution = Solution(problem)
    solution.objectives[:] = [float(x) for x in args]
    return solution

class TestSolution(unittest.TestCase):
    
    def test_deepcopy(self):
        orig = createSolution(4, 5)
        orig.constraint_violation = 2
        orig.evaluated = True
        
        clone = copy.deepcopy(orig)
        
        self.assertEqual(orig.problem, clone.problem)
        self.assertEqual(4, clone.objectives[0])
        self.assertEqual(5, clone.objectives[1])
        self.assertEqual(2, clone.constraint_violation)
        self.assertEqual(True, clone.evaluated)
        

class TestConstraint(unittest.TestCase):
    
    def test_eq(self):
        constraint = Constraint("==0")
        self.assertEqual(0.0, constraint(0.0))
        self.assertNotEqual(0.0, constraint(1.0))
        self.assertNotEqual(0.0, constraint(-1.0))
        
        constraint = Constraint("==5")
        self.assertEqual(0.0, constraint(5.0))
        self.assertNotEqual(0.0, constraint(-5.0))
        self.assertNotEqual(0.0, constraint(10.0))
        
    def test_lte(self):
        constraint = Constraint("<=0")
        self.assertEqual(0.0, constraint(0.0))
        self.assertEqual(0.0, constraint(-1.0))
        self.assertNotEqual(0.0, constraint(1.0))
        
    def test_gte(self):
        constraint = Constraint(">=0")
        self.assertEqual(0.0, constraint(0.0))
        self.assertEqual(0.0, constraint(1.0))
        self.assertNotEqual(0.0, constraint(-1.0))
        
    def test_lt(self):
        constraint = Constraint("<0")
        self.assertNotEqual(0.0, constraint(0.0))
        self.assertNotEqual(0.0, constraint(1.0))
        self.assertEqual(0.0, constraint(-1.0))
        
    def test_gt(self):
        constraint = Constraint(">0")
        self.assertNotEqual(0.0, constraint(0.0))
        self.assertEqual(0.0, constraint(1.0))
        self.assertNotEqual(0.0, constraint(-1.0))
        
    def test_neq(self):
        constraint = Constraint("!=0")
        self.assertNotEqual(0.0, constraint(0.0))
        self.assertEqual(0.0, constraint(1.0))
        self.assertEqual(0.0, constraint(-1.0))
        
class TestParetoDominance(unittest.TestCase):
    
    def test_dominance(self):
        dominance = ParetoDominance()
        s1 = createSolution(0.0, 0.0)
        s2 = createSolution(1.0, 1.0)
        s3 = createSolution(0.0, 1.0)
        
        self.assertEqual(-1, dominance.compare(s1, s2))
        self.assertEqual(1, dominance.compare(s2, s1))
        self.assertEqual(-1, dominance.compare(s1, s3))
        self.assertEqual(1, dominance.compare(s3, s1))
        
    def test_nondominance(self):
        dominance = ParetoDominance()
        s1 = createSolution(0.0, 1.0)
        s2 = createSolution(0.5, 0.5)
        s3 = createSolution(1.0, 0.0)
        
        self.assertEqual(0, dominance.compare(s1, s2))
        self.assertEqual(0, dominance.compare(s2, s1))
        self.assertEqual(0, dominance.compare(s2, s3))
        self.assertEqual(0, dominance.compare(s3, s2))
        self.assertEqual(0, dominance.compare(s1, s3))
        self.assertEqual(0, dominance.compare(s3, s1))
        
class TestArchive(unittest.TestCase):
    
    def test_dominance(self):
        s1 = createSolution(1.0, 1.0)
        s2 = createSolution(0.0, 0.0)
        s3 = createSolution(0.0, 1.0)
        
        archive = Archive(ParetoDominance())
        archive += [s1, s2, s3]
        
        self.assertEqual(1, len(archive))
        self.assertEqual(s2, archive[0])
        
    def test_nondominance(self):
        s1 = createSolution(0.0, 1.0)
        s2 = createSolution(0.5, 0.5)
        s3 = createSolution(1.0, 0.0)
        
        archive = Archive(ParetoDominance())
        archive += [s1, s2, s3]
        
        self.assertEqual(3, len(archive))
        
class TestNondominatedSort(unittest.TestCase):
    
    def setUp(self):
        self.s1 = createSolution(0.0, 1.0)
        self.s2 = createSolution(0.5, 0.5)
        self.s3 = createSolution(1.0, 0.0)
        self.s4 = createSolution(0.75, 0.75)
        self.s5 = createSolution(1.0, 1.0)
        
        self.population = [self.s1, self.s2, self.s3, self.s4, self.s5]
        random.shuffle(self.population)
        
    def test_ranking(self):
        nondominated_sort(self.population)
        
        self.assertEqual(0, self.s1.rank)
        self.assertEqual(0, self.s2.rank)
        self.assertEqual(0, self.s3.rank)
        self.assertEqual(1, self.s4.rank)
        self.assertEqual(2, self.s5.rank)
        
    def test_crowding(self):
        nondominated_sort(self.population)
        
        self.assertEqual(POSITIVE_INFINITY, self.s1.crowding_distance)
        self.assertEqual(2, self.s2.crowding_distance)
        self.assertEqual(POSITIVE_INFINITY, self.s3.crowding_distance)
        self.assertEqual(POSITIVE_INFINITY, self.s4.crowding_distance)
        self.assertEqual(POSITIVE_INFINITY, self.s5.crowding_distance)
        
    def test_split2(self):
        nondominated_sort(self.population)
        (first, second) = nondominated_split(self.population, 2)
        
        self.assertEqual(0, len(first))
        self.assertEqual(3, len(second))
        self.assertIn(self.s1, second)
        self.assertIn(self.s2, second)
        self.assertIn(self.s3, second)
        
    def test_split3(self):
        nondominated_sort(self.population)
        (first, second) = nondominated_split(self.population, 3)
        
        self.assertEqual(3, len(first))
        self.assertEqual(0, len(second))
        self.assertIn(self.s1, first)
        self.assertIn(self.s2, first)
        self.assertIn(self.s3, first)
        
    def test_split4(self):
        nondominated_sort(self.population)
        (first, second) = nondominated_split(self.population, 4)
        
        self.assertEqual(4, len(first))
        self.assertEqual(0, len(second))
        self.assertIn(self.s1, first)
        self.assertIn(self.s2, first)
        self.assertIn(self.s3, first)
        self.assertIn(self.s4, first)
        
    def test_truncate2(self):
        nondominated_sort(self.population)
        result = nondominated_truncate(self.population, 2)
        
        self.assertEqual(2, len(result))
        self.assertIn(self.s1, result)
        self.assertIn(self.s3, result)
        
    def test_truncate4(self):
        nondominated_sort(self.population)
        result = nondominated_truncate(self.population, 4)
        
        self.assertEqual(4, len(result))
        self.assertIn(self.s1, result)
        self.assertIn(self.s2, result)
        self.assertIn(self.s3, result)
        self.assertIn(self.s4, result)
        
    def test_prune2(self):
        nondominated_sort(self.population)
        result = nondominated_prune(self.population, 2)
        
        self.assertEqual(2, len(result))
        self.assertIn(self.s1, result)
        self.assertIn(self.s3, result)
        
    def test_prune4(self):
        nondominated_sort(self.population)
        result = nondominated_prune(self.population, 4)
        
        self.assertEqual(4, len(result))
        self.assertIn(self.s1, result)
        self.assertIn(self.s2, result)
        self.assertIn(self.s3, result)
        self.assertIn(self.s4, result)
        
    def test_truncate_fitness_max(self):
        self.s1.fitness = 1
        self.s2.fitness = 5
        self.s3.fitness = 3
        self.s4.fitness = 4
        self.s5.fitness = 2
        
        result = truncate_fitness(self.population, 3)
        
        self.assertEqual(3, len(result))
        self.assertIn(self.s2, result)
        self.assertIn(self.s4, result)
        self.assertIn(self.s3, result)
        
    def test_truncate_fitness_min(self):
        self.s1.fitness = 1
        self.s2.fitness = 5
        self.s3.fitness = 3
        self.s4.fitness = 4
        self.s5.fitness = 2
        
        result = truncate_fitness(self.population, 3, larger_preferred=False)
        
        self.assertEqual(3, len(result))
        self.assertIn(self.s1, result)
        self.assertIn(self.s5, result)
        self.assertIn(self.s3, result)
        
class TestNormalize(unittest.TestCase):
    
    def test_normalize(self):
        s1 = createSolution(0, 2)
        s2 = createSolution(2, 3)
        s3 = createSolution(1, 1)
        solutions = [s1, s2, s3]
        
        normalize(solutions)
        
        self.assertEqual([0.0, 0.5], s1.normalized_objectives)
        self.assertEqual([1.0, 1.0], s2.normalized_objectives)
        self.assertEqual([0.5, 0.0], s3.normalized_objectives)
        
class TestEpsilonBoxArchive(unittest.TestCase):
    
    def test_improvements(self):
        s1 = createSolution(0.25, 0.25)
        s2 = createSolution(0.1, 0.1)
        s3 = createSolution(0.245, 0.245)
        s4 = createSolution(0.1, 0.5)
        s5 = createSolution(0.5, 0.5)
        s6 = createSolution(0.0, 0.0)
        
        archive = EpsilonBoxArchive([0.1])
        
        archive.extend([s1, s2, s3, s4, s5, s6])
        self.assertEqual(2, archive.improvements)


from .tools import *
from .commonImport import *
from .collection import *
from .color_palette import *
from .graphs import *
from .graphs3 import *
from .linkDataGraph import *
from .options import *
from .additional_tools import *
from .myjson import *
from .tikzTranslator import export_to_tikz_groupGraphs, export_to_tikz_contour_plot
from .inkscape_manager import *



from abc import ABCMeta, abstractmethod


class DeviceDrawerInterface(metaclass=ABCMeta):
    @abstractmethod
    def draw(self, theDevice):
        pass

    @abstractmethod
    def get_init_camera(self, theDevice):
        pass

    def keyboard_push_action(self, theKey):
        pass

    def get_colour_scalebar(self):
        return [0.8, 0.84, 0.9]

    def get_colour_background(self):
        return [0.2, 0.2, 0.2]

    def get_opengl_options(self):
        pass



from .onclickInterface import OnclickInterface
from PyQt5 import QtWidgets
import os


class Onclick_exportTrace(OnclickInterface):
    """On click: export the data of the whole the trace selected"""
    def __init__(self, theDataLink, getShadow=True):
        """

        :param theDataLink: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        """
        self.theDataLink = theDataLink
        self.getShadow = getShadow

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        theCollection = self.theDataLink.get_collection_from_graph(index_graph, index_trace, getShadow=self.getShadow)
        dlg = QtWidgets.QFileDialog.getSaveFileName()[0]
        if dlg:
            root, ext = os.path.splitext(dlg)
            filename_collection = root + theCollection.get_extension()
            theCollection.clone(filename_collection)

    def get_name(self):
        return "Export collection of the trace"



import math
from optimeed.core import printIfShown, SHOW_WARNING

try:
    import scipy.spatial.qhull as qhull
    has_scipy = True
except ImportError:
    has_scipy = False
import numpy as np


class fast_LUT_interpolation:
    """Class designed for fast interpolation in look-up table when successive searchs are called often.
    Otherwise use griddata"""

    def __init__(self, independent_variables, dependent_variables):
        """

        :param independent_variables: np.array of shape (N,d) containing the independent variables of the dataset to interpolate. Typically X for 1D LU, or XY for 2D LUT
        :param dependent_variables: np.array of shape (1,d) containing the dependet variables of the dataset to interpolate. Typically Y for 1D LU, or Z for 2D LUT
        """
        self.param_ind = independent_variables
        self.param_dep = dependent_variables
        self.tri = self.interp_tri(independent_variables)
        self.dim = independent_variables.shape[1]

    @staticmethod
    def interp_tri(xyz):
        if has_scipy:
            tri = qhull.Delaunay(xyz)
        else:
            printIfShown("Scipy library not found -> fast_LUT_interpolation has been deactivated", SHOW_WARNING)
        return tri

    def interpolate(self, point, fill_value=np.nan):
        """
        Perform the interpolation
        :param point: coordinates to interpolate (tuple or list of tuples for multipoints)
        :param fill_value: value to put if extrapolated.
        :return: coordinates
        """
        d = self.dim
        tri = self.tri
        values = self.param_dep
        simplex = tri.find_simplex(point)
        vertices = np.take(tri.simplices, simplex, axis=0)
        temp = np.take(tri.transform, simplex, axis=0)
        delta = point - temp[:, d]
        bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)
        wts = np.hstack((bary, 1.0 - bary.sum(axis=1, keepdims=True)))
        ret = np.einsum('nj,nj->n', np.take(values, vertices), wts)
        ret[np.any(wts < 0, axis=1)] = fill_value
        return ret


def interpolate_table(x0, x_values, y_values):
    """From sorted table (x,y) find y0 corresponding to x0 (linear interpolation)"""
    index = np.searchsorted(x_values, x0)
    def _interpolate(x0, x_values, y_values, index):
        if index == len(x_values):
            return y_values[-1]
        if index == 0:
            return y_values[0]
        x_before, y_before = x_values[index-1], y_values[index-1]
        x_after, y_after = x_values[index], y_values[index]

        for x, y in [(x_before, y_before), (x_after, y_after)]:
            if x == x0:
                return y

        ratio = (x0 - x_before)/(x_after-x_before)
        return y_before + ratio*(y_after - y_before)

    if isinstance(index, np.ndarray):
        results = [0.0]*len(index)
        for k, ind in enumerate(index):
            results[k] = _interpolate(x0[k], x_values, y_values, ind)
        return results
    else:
        return _interpolate(x0, x_values, y_values, index)


def derivate(t, y):
    derivated = [0.0]*len(y)
    for i in range(1, len(y)-1):
        derivated[i] = (y[i+1] - y[i-1])/(t[i+1] - t[i-1])
    derivated[-1] = (y[1] - y[-2])/((t[-1]-t[-2]) + (t[1]-t[0]))
    derivated[0] = derivated[-1]
    return derivated


def linspace(start, stop, npoints):
    return np.linspace(start, stop, npoints).tolist()


def reconstitute_signal(amplitudes, phases, numberOfPeriods=1, x_points=None, n_points=50):
    """Reconstitute the signal from fft. Number of periods of the signal must be specified if different of 1"""
    if x_points is None:
        x_points = np.linspace(0, 2*np.pi*numberOfPeriods, num=int(n_points))
    n_points = len(x_points)
    y = np.zeros(n_points)
    for key in amplitudes:
        y += amplitudes[key] * np.cos(key/numberOfPeriods * x_points + phases[key])
    return x_points, y
# def my_ifft(amplitude, angle, hasBeenTruncated=True):
#     """Inverse fft of signal.
#     If hasBeenTruncated (initial fft signal was full period (including first point of next period):
#     Axis vector must be multiplied by (n_points-1)/(n_points-2)"""
#     amplitudes = list(amplitude.values())
#     angles = list(angle.values())
#     n_points = (len(angles))*2
#     normalization = 2/n_points
#     if hasBeenTruncated:
#         normalization *= (n_points+1)/(n_points-1)
#     coeffs = np.array([amplitudes[i] * np.exp(1j*angles[i]) for i in range(len(angles))])
#     return np.fft.irfft(coeffs)/normalization


def my_fft(y):
    """Real FFT of signal Bx, with real amplitude of harmonics. Input signal must be within a period."""
    if np.isclose(y[0], y[-1]):
        y = y[:-1]
    n_points = len(y)
    normalization = 2/n_points
    res = np.fft.rfft(y)*normalization
    res[0] /= 2

    amplitudes = np.abs(res)
    phases = np.angle(res)
    max_amplitude = max(amplitudes[1:])
    amplitude = dict()
    phase = dict()

    for harmonic in range(len(amplitudes)):
        if harmonic > 0 and amplitudes[harmonic] > 0.01*max_amplitude:
            amplitude[harmonic] = amplitudes[harmonic]
            phase[harmonic] = phases[harmonic]

    # harmonics = list(range(len(res)))
    # amplitude = dict(zip(harmonics, np.abs(res)))
    # phase = dict(zip(harmonics, np.angle(res)))
    return amplitude, phase


def cart2pol(x, y):
    rho = np.sqrt(x ** 2 + y ** 2)
    phi = np.arctan2(y, x)
    return rho, phi


def pol2cart(rho, phi):
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return x, y


def partition(array, begin, end):
    pivot = begin
    for i in range(begin+1, end+1):
        if array[i] <= array[begin]:
            pivot += 1
            array[i], array[pivot] = array[pivot], array[i]
    array[pivot], array[begin] = array[begin], array[pivot]
    return pivot


def quicksort(array):
    end = len(array) - 1

    def _quicksort(_array, _begin, _end):
        if _begin >= _end:
            return
        pivot = partition(_array, _begin, _end)
        _quicksort(_array, _begin, pivot-1)
        _quicksort(_array, pivot+1, _end)
    _quicksort(array, 0, end)
    return array


def dist(p, q):
    """Return the Euclidean distance between points p and q.
    :param p: [x, y]
    :param q: [x, y]
    :return: distance (float)"""
    return math.hypot(p[0] - q[0], p[1] - q[1])


def sparse_subset(points, r):
    """Returns a maximal list of elements of points such that no pairs of
    points in the result have distance less than r.
    :param points: list of tuples (x,y)
    :param r: distance
    :return: corresponding subset (list), indices of the subset (list)"""
    result = []
    indices = []
    for i, p in enumerate(points):
        if all(dist(p, q) >= r for q in result):
            result.append(p)
            indices.append(i)
    return result, indices


def integrate(x, y):
    """
    Performs Integral(x[0] to x[-1]) of y dx

    :param x: x axis coordinates (list)
    :param y: y axis coordinates (list)
    :return: integral value
    """
    integral = 0
    for i in range(1, len(x)):
        dx = x[i]-x[i-1]
        integral += dx*(y[i] + y[i-1])/2
    return integral


def my_fourier(x, y, n, L):
    """
    Fourier analys

    :param x: x axis coordinates
    :param y: y axis coordinates
    :param n: number of considered harmonic
    :param L: half-period length
    :return: a and b coefficients (y = a*cos(x) + b*sin(y))
    """
    y_to_integrate_a = [None]*len(y)
    y_to_integrate_b = [None]*len(y)

    for i in range(len(y)):
        y_to_integrate_a[i] = y[i] * np.cos(n*np.pi*x[i] / L)
        y_to_integrate_b[i] = y[i] * np.sin(n*np.pi*x[i] / L)

    a_n = 1/L * integrate(x, y_to_integrate_a)
    b_n = 1/L * integrate(x, y_to_integrate_b)
    return a_n, b_n


def get_ellipse_axes(a, b, dphi):
    """Trouve les longueurs des axes majeurs et mineurs de l'ellipse, ainsi que l'orientation de l'ellipse.
    ellipse: x(t) = A*cos(t), y(t) = B*cos(t+dphi)
    Etapes: longueur demi ellipse CENTRéE = sqrt(a^2 cos^2(x) + b^2 cos^2(t+phi)
    Minimisation de cette formule => obtention formule tg(2x) = alpha/beta"""
    assert a >= 0, 'a must be positive'
    assert b >= 0, 'b must be positive'

    if np.isclose(dphi, 0):  # Line
        if a == 0:
            phase = np.pi/2
        else:
            phase = np.arctan(b/a)
        return np.sqrt(a**2 + b**2), 0.0, phase
    elif np.isclose(dphi, np.pi/2):  # Properly oriented ellipse
        if a > b:
            return a, b, 0.0
        return b, a, np.pi/2
    sdphi = np.sin(dphi)
    cdphi = np.cos(dphi)
    alpha = (a**2)/2 + (b**2)/2*(cdphi**2 - sdphi**2)
    beta = -b**2*(sdphi*cdphi)
    t_optim_1, t_optim_2 = 0.5*(np.arctan(beta/alpha) + np.pi), 0.5*(np.arctan(beta/alpha) + 2*np.pi)
    max_length = np.sqrt((a * np.cos(t_optim_1))**2 + (b * np.cos(t_optim_1 + dphi))**2)
    min_length = np.sqrt((a * np.cos(t_optim_2))**2 + (b * np.cos(t_optim_2 + dphi))**2)
    # print(np.cos(t_optim_1 + dphi)/np.cos(t_optim_1), np.cos(t_optim_2 + dphi)/np.cos(t_optim_2))
    if max_length > min_length:
        phase = np.arctan(b/a * np.cos(t_optim_1 + dphi)/np.cos(t_optim_1))
    else:
        phase = np.arctan(b/a * np.cos(t_optim_2 + dphi)/np.cos(t_optim_2))
        max_length, min_length = min_length, max_length

    return max_length, min_length, phase


def convert_color(color):
    """
    Convert a color to a tuple if color is a char, otherwise return the tuple.

    :param color: (r,g,b) or char.
    :return:
    """
    try:
        colors = {"b": (0, 0, 255), "r": (255, 0, 0), "g": (0, 255, 0), "y": (255, 255, 0), "c": (0, 255, 255), "k": (0, 0, 0), "w": (255, 255, 255), "m": (255, 0, 255), "o": (255, 120, 0)}
        return colors[color]
    except KeyError:
        if type(color) == list or type(color) == tuple:
            return color[0], color[1], color[2]
        printIfShown("Unknown color {}, returning blue".format(color), SHOW_WARNING)
        return 0, 0, 255


def convert_color_with_alpha(color, alpha=255):
    """
    Same as meth:`convert_color` but with transparency
    """
    r, g, b = convert_color(color)
    return r, g, b, alpha




from PyQt5 import QtCore, QtGui, QtWidgets


class Widget_lineDrawer(QtWidgets.QWidget):
    """Widget allowing to display several lines easily"""

    signal_must_update = QtCore.pyqtSignal(object)

    # listOfLines[i] = [x_1, y_1, x_2, y_2]
    def __init__(self, minWinHeight=300, minWinWidth=300, is_light=True):
        super().__init__()
        # Set windows width
        self.setMinimumHeight(minWinHeight)
        self.setMinimumWidth(minWinWidth)

        # Set background colour
        p = self.palette()
        self.is_light = is_light
        if self.is_light:
            p.setColor(self.backgroundRole(), QtCore.Qt.white)
        else:
            p.setColor(self.backgroundRole(), QtCore.Qt.black)
        self.setPalette(p)
        self.setAutoFillBackground(True)

        # Set objects to draw
        self.dict_listOfLines = {}
        self.dict_pens = {}

        self.signal_must_update.connect(self.on_update_signal)

    def on_update_signal(self, listOfLines):
        self.set_lines(listOfLines=listOfLines)

    def delete_lines(self, key_id):
        """
        Dele the lines
        :param key_id: id to delete
        :return:
        """
        del self.dict_listOfLines[key_id]
        del self.dict_pens[key_id]

    def set_lines(self, listOfLines, key_id=0, pen=None):
        """
        Set the lines to display
        :param listOfLines: list of [x1, y1, x2, y2] corresponding to lines
        :param key_id: id of the trace
        :param pen: pen used to draw the lines
        :return:
        """
        for k, line in enumerate(listOfLines):
            if len(line) == 6:  # Coordinates are (x,y,z) and not (x,y) -> compress
                listOfLines[k] = [line[0], line[1], line[3], line[4]]

        self.dict_listOfLines[key_id] = listOfLines
        if pen is None:
            if self.is_light:
                self.dict_pens[key_id] = QtGui.QPen(QtCore.Qt.black, 1)
            else:
                self.dict_pens[key_id] = QtGui.QPen(QtCore.Qt.white, 1)
        else:
            self.dict_pens[key_id] = QtGui.QPen(pen.color())  # pen
        self.repaint()
        self.update()

    def paintEvent(self, event, painter=None):
        if painter is None:
            painter = QtGui.QPainter(self)
            painter.setRenderHint(QtGui.QPainter.Antialiasing)

        x_min, x_max, y_min, y_max = self.get_extrema_lines()
        min_margin = 20
        maxwidth = x_max - x_min
        maxheight = y_max - y_min
        width_window = self.rect().width()
        height_window = self.rect().height()

        try:
            ratio_width = (width_window - min_margin * 2) / maxwidth
            ratio_height = (height_window - min_margin * 2) / maxheight
        except ZeroDivisionError:
            ratio_width = 1
            ratio_height = 1
        ratio = min(ratio_width, ratio_height)

        center_x = (width_window + min_margin * 2 - ratio * maxwidth) / 2
        painter.translate(-min_margin + center_x - x_min * ratio, -(min_margin - y_min * ratio) + height_window)

        for key in self.dict_listOfLines:
            # Set painters and objects to draw
            thePen = self.dict_pens[key]
            thePen.setWidthF(2)
            painter.setPen(thePen)

            for line in self.dict_listOfLines[key]:
                painter.drawLine(line[0]*ratio, -line[1]*ratio, line[2]*ratio, -line[3]*ratio)

            thePen.setWidthF(0.5)
            painter.setPen(thePen)
            painter.drawLine(0, 0, x_max * ratio, 0)
            painter.drawLine(0, 0, 0, -y_max * ratio)

    def get_extrema_lines(self):
        x_min, x_max, y_min, y_max = 0, 1e-12, 0, 1e-12
        for i, key in enumerate(self.dict_listOfLines):
            try:
                x_min_temp = min(min(x[0], x[2]) for x in self.dict_listOfLines[key])
                x_max_temp = max(max(x[0], x[2]) for x in self.dict_listOfLines[key])
                y_min_temp = min(min(x[1], x[3]) for x in self.dict_listOfLines[key])
                y_max_temp = max(max(x[1], x[3]) for x in self.dict_listOfLines[key])

                if i == 0:
                    x_min = x_min_temp
                    x_max = x_max_temp
                    y_min = y_min_temp
                    y_max = y_max_temp
                else:
                    x_min = min(x_min, x_min_temp)
                    x_max = max(x_max, x_max_temp)
                    y_min = min(y_min, y_min_temp)
                    y_max = max(y_max, y_max_temp)
            except (IndexError, ValueError):
                pass
        return x_min, x_max, y_min, y_max




from optimeed.core.tools import applyEquation
from .interfaceObjCons import InterfaceObjCons


class FastObjCons(InterfaceObjCons):
    """Convenience class to create an objective or a constraint very fast."""
    def __init__(self, constraintEquation, name=None):
        super().__init__()
        self.constraintEquation = constraintEquation
        self.name = name

    def compute(self, theDevice):
        return applyEquation(theDevice, self.constraintEquation)

    def get_name(self):
        if self.name is None:
            return self.constraintEquation
        return str(self.name)



import os

import numpy as np

from .graphs import Graphs, Graph, Data
from .graphs3 import SurfPlot, ContourPlot, FilledContourPlot, MeshPlot, ScatterPlot3, GridPlot_Generic
from .tools import indentParagraph

templates_tikz = os.path.join(os.path.dirname(__file__), 'templates_tikz')


def format_escape_char(theStr):
    theStr = theStr.replace("%", r"\%")
    theStr = theStr.replace("_", r"\_")
    return theStr


def convert_linestyle(linestyle):
    try:
        linestyles = {"-": '', "--": "dashed", "..": "dotted", ".": "dotted"}
        return linestyles[linestyle]
    except KeyError:
        return linestyle


def find_all_colors(theGraphs):
    all_colors = dict()
    correspondances = dict()
    for graphId in theGraphs.get_all_graphs_ids():
        correspondances[graphId] = dict()
        for traceId in theGraphs.get_graph(graphId).get_all_traces():
            try:
                r, g, b = theGraphs.get_graph(graphId).get_trace(traceId).get_color()
            except TypeError:
                r, g, b = 0, 0, 0
            key_format = "RGB{}_{}_{}".format(r, g, b)
            all_colors[key_format] = r"\definecolor{{{}}}{{RGB}}{{{},{},{}}}".format(key_format, r, g, b)
            correspondances[graphId][traceId] = key_format
    return all_colors, correspondances


def convert_marker(marker):
    try:
        markers = {"o": "*", "t1": "triangle", "t2": "triangle", "t3": "triangle", "t": "triangle", "s": "square", "d": "diamond", "p": "pentagon"}
        return markers[marker]
    except KeyError:
        return marker


def do_preamble():
    with open(os.path.join(templates_tikz, "preamble.tex"), "r") as f:
        return f.read()


def do_generate_figure():
    with open(os.path.join(templates_tikz, "generate_figure.tex"), "r") as f:
        return f.read()


def do_specific_axis_options(theGraph: Graph):
    """Get graph-specific axis options"""

    theStr = ''
    for key in theGraph.get_all_traces():
        xlabel = theGraph.get_trace(key).get_x_label()
        if xlabel:
            theStr += "xlabel={{ {} }},".format(format_escape_char(xlabel))
            break
    for key in theGraph.get_all_traces():
        ylabel = theGraph.get_trace(key).get_y_label()
        if ylabel:
            theStr += "ylabel={{ {} }},".format(format_escape_char(ylabel))
            break
    return theStr


def do_specific_trace_options(theTrace: Data, theColor):
    """Get latex trace options from Data"""
    linestyle = convert_linestyle(theTrace.get_linestyle())
    is_filled = theTrace.symbol_isfilled()
    markerstyle = convert_marker(theTrace.get_symbol())
    is_scattered = theTrace.isScattered
    outline = theTrace.get_symbolOutline()

    theStr = '\ncolor={},'.format(theColor)
    if is_scattered:
        theStr += "\nonly marks,"
    else:
        theStr += "{},".format(linestyle)
    theStr += '\nmark={},'.format(markerstyle)
    mark_options = '\nsolid, \nline width=0.1pt, \nmark size=1pt, \ndraw={}!{}!{},'.format(theColor, int(abs(outline - 2) * 100), "white" if outline < 1 else "black")
    if not is_filled:
        mark_options += "\nfill opacity=0,\n"
    theStr += "\nmark options={{{}}}".format(indentParagraph(mark_options, 1))
    return theStr


def export_to_tikz_groupGraphs(theGraphs: Graphs, foldername, additionalPreamble=lambda: '', additionalAxisOptions=lambda graphId: '', additionalTraceOptions=lambda graphId, traceId: '', debug=False):
    """
    Export the graphs as group

    :param theGraphs: Graphs to save
    :param foldername: Foldername to save
    :param additionalPreamble: method that returns string for custom tikz options
    :param additionalAxisOptions: method that returns string for custom tikz options
    :param additionalTraceOptions: method that returns string for custom tikz options
    :return:
    """
    all_colors, correspondance_colors = find_all_colors(theGraphs)

    if not debug:
        if not foldername:
            return
        os.makedirs(foldername, exist_ok=True)
        source_folder = os.path.join(foldername, "source")
        os.makedirs(source_folder, exist_ok=True)

    # Data file creation
    if not debug:
        for graphId in theGraphs.get_all_graphs_ids():
            theGraph = theGraphs.get_graph(graphId)
            for traceId in theGraph.get_all_traces():
                filename = os.path.join(source_folder, "g{}t{}.dat".format(graphId, traceId))
                theTrace = theGraph.get_trace(traceId)
                x, y = theTrace.get_plot_data()
                meta = theTrace.get_meta()
                if meta is not None:
                    with open(filename, "w") as f:
                        f.write("x\ty\tmeta\n")
                        for i in range(len(x)):
                            f.write("{}\t{}\t{}\n".format(x[i], y[i], meta[i]))
                else:
                    np.savetxt(filename, np.array([[x[i], y[i]] for i in range(len(x))]), header='x\ty')
                # add_body += "\\pgfplotstableread{{source/g{}t{}.dat}}{{\\g{}t{}}}\n".format(graphId, traceId, graphId, traceId)

    # Preamble
    theStr = ''
    theStr += do_preamble() + additionalPreamble()
    for color in all_colors.values():
        theStr += "% Colors\n{}\n".format(color)
    if not debug:
        with open(os.path.join(foldername, "preamble.tex"), "w") as f:
            f.write(theStr)
    else:
        print("PREAMBLE\n{}\n\n\n".format(theStr))
    # Generate picture creation
    theStr = ''
    theStr += do_generate_figure()
    if not debug:
        with open(os.path.join(foldername, "generate_figure.tex"), "w") as f:
            f.write(theStr)
    else:
        print("GENERATOR\n{}\n\n\n".format(theStr))

    # picture file creation
    theStr = ''
    theStr += indentParagraph("\\begin{tikzpicture}\n", 1)
    groupStyle = "group style={{group size=1 by {},}},".format(len(theGraphs.get_all_graphs_ids()))
    theStr += indentParagraph("\\begin{{groupplot}}[default_axis_style, {}]".format(groupStyle), 3)

    for graphId in theGraphs.get_all_graphs_ids():
        theGraph = theGraphs.get_graph(graphId)
        theStr += indentParagraph("\\nextgroupplot[{}]".format(do_specific_axis_options(theGraph) + additionalAxisOptions(graphId)), 5)
        for traceId in theGraph.get_all_traces():
            theTrace = theGraph.get_trace(traceId)
            filename = "source/g{}t{}.dat".format(graphId, traceId)
            theStr += indentParagraph("\\addplot[default_trace_style, {}] file {{\\relativepath {}}};\n".format(indentParagraph(do_specific_trace_options(theTrace, correspondance_colors[graphId][traceId]) + additionalTraceOptions(graphId, traceId), 3), filename), 7)
            if theTrace.get_legend():
                theStr += indentParagraph("\\addlegendentry{{{}}}".format(theTrace.get_legend()), 7)
    theStr += indentParagraph("\\end{groupplot}\n", 2)
    theStr += indentParagraph("\\end{tikzpicture}\n", 1)
    if not debug:
        with open(os.path.join(foldername, "figure.tex"), "w") as f:
            f.write(theStr)
    else:
        print("FIGURE\n{}\n\n\n".format(theStr))


""" Content from here is destined to 3D plots"""


def do_preamble3D():
    with open(os.path.join(templates_tikz, "preamble3D.tex"), "r") as f:
        return f.read()


def format_Griddata(X, Y, Z):
    theStr = ''
    for row in range(len(Z)):
        for col in range(len(Z[row])):
            theStr += "{}\t{}\t{}\n".format(X[col], Y[row], Z[row, col])
        theStr += "\n"
    return theStr


def format_scatterdata(x, y, z):
    theStr = ''
    for i in range(len(x)):
        theStr += "{}\t{}\t{}\n".format(x[i], y[i], z[i])
    return theStr


def export_to_tikz_contour_plot(list_of_traces3, foldername, filename_data="data"):
    """
    Export the graphs as group

    :param list_of_traces3: List of 3D traces
    :param foldername: Foldername to save
    :param filename_data: filename of the data
    :return:
    """
    if not foldername:
        return
    os.makedirs(foldername, exist_ok=True)
    source_folder = os.path.join(foldername, "source")
    os.makedirs(source_folder, exist_ok=True)

    # Data file creation
    for k, trace in enumerate(list_of_traces3):
        filename = os.path.join(source_folder, "{}_{}.dat".format(filename_data, k))
        with open(filename, 'w', encoding="utf-8") as f:
            if isinstance(trace, GridPlot_Generic):
                f.write(format_Griddata(*trace.get_plot_data()))
            else:
                f.write(format_scatterdata(*trace.get_plot_data()))

    # Preamble
    theStr = ''
    theStr += do_preamble3D()
    with open(os.path.join(foldername, "preamble.tex"), "w") as f:
        f.write(theStr)
    # Generate picture creation
    theStr = ''
    theStr += do_generate_figure()
    with open(os.path.join(foldername, "generate_figure.tex"), "w") as f:
        f.write(theStr)

    # picture file creation
    default_view = '{0}{90}' if any(map(lambda tr: isinstance(tr, ContourPlot) or isinstance(tr, FilledContourPlot), list_of_traces3)) else '{35}{55}'
    theStr = ''
    theStr += indentParagraph("\\begin{tikzpicture}\n", 1)
    theStr += indentParagraph("\\begin{{axis}}[default_axis_style, view = {}, grid on]\n".format(default_view), 2)
    for k, trace in enumerate(list_of_traces3):
        args = ''
        add_line = ''
        if isinstance(trace, SurfPlot):
            args = 'surf, line width = 0.1pt, unbounded coords=jump'
        elif isinstance(trace, MeshPlot):
            args = 'mesh, scatter, mark size = 1 pt, line width = 0.1pt, unbounded coords=jump'
        elif isinstance(trace, FilledContourPlot):
            levels = '{' + ','.join([str(lev) for lev in trace.get_levels()]) + '}'
            args = 'contour gnuplot = {{number={},labels={{true}}, levels={}, draw color=black, labels over line}}, unbounded coords=jump'.format(trace.get_number_of_contours(), levels)
            add_line = '\\addplot3[surf, shader=interp, unbounded coords=jump] file {{\\relativepath {}}};\n'.format("source/{}_{}.dat".format(filename_data, k))
        elif isinstance(trace, ContourPlot):
            levels = '{' + ','.join([str(lev) for lev in trace.get_levels()]) + '}'
            args = 'contour gnuplot = {{number={},labels={{true}}, levels={}}}, unbounded coords=jump'.format(trace.get_number_of_contours(), levels)
        elif isinstance(trace, ScatterPlot3):
            args = 'scatter, only marks, mark size=1.5pt, scatter/use mapped color={fill=mapped color, draw=black}, unbounded coords=jump'
        theStr += indentParagraph("{}\n".format(add_line), 3)
        theStr += indentParagraph("\\addplot3[{}] file {{\\relativepath {}}};\n".format(args, "source/{}_{}.dat".format(filename_data, k)), 3)
    theStr += indentParagraph("\\end{axis}\n", 2)
    theStr += indentParagraph("\\end{tikzpicture}\n", 1)
    with open(os.path.join(foldername, "figure.tex"), "w") as f:
        f.write(theStr)



from .pyqtgraph.graphicsItems.LegendItem import ItemSample
from .pyqtgraph.graphicsItems.ScatterPlotItem import ScatterPlotItem, drawSymbol
from .pyqtgraph.graphicsItems.PlotDataItem import PlotDataItem
from .pyqtgraph import functions as fn
from optimeed.visualize.graphs import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets, QtOpenGL
import os

isOnWindows = os.name == 'nt'


"""Other modified files (directly): ScatterPlotItem.py, to change point selection. Ctrl + clic: select area. Clic: only one single point:


class OnClicSelector:
    def __init__(self):
        self.p_list = []

    def add_point(self, newp):
        self.p_list.append(newp)

    def draw(self, painter):
        if len(self.p_list) > 2:
            pen = fn.mkPen(1)
            pen.setWidthF(2)
            painter.setPen(pen)
            painter.drawPolyline(QtGui.QPolygonF(self.p_list))

    def reset(self):
        self.p_list = []

    def getPath(self):
        return path.Path([(p.x(), p.y()) for p in self.p_list] + [(self.p_list[-1].x(), self.p_list[-1].y())])

------------------------------
    def mouseDragEvent(self, ev):
        if ev.modifiers() and QtCore.Qt.ControlModifier:
            ev.accept()
            self.clicSelector.add_point(ev.pos())
            if ev.isFinish():
                path = self.clicSelector.getPath()
                points = self.points()
                contains_points = path.contains_points([(p.pos().x(), p.pos().y()) for p in points])
                indices = [i for i, cond in enumerate(contains_points) if cond]
                points_clicked = [points[i] for i in indices]
                self.ptsClicked = points_clicked
                self.sigClicked.emit(self, self.ptsClicked)
                self.clicSelector.reset()
            self.update()
        else:
            ev.ignore()
"""


class myGraphicsLayoutWidget(pg.GraphicsView):
    def __init__(self, parent=None, **_kwargs):
        pg.GraphicsView.__init__(self, parent)
        self.ci = myGraphicsLayout()
        for n in ['nextRow', 'nextCol', 'nextColumn', 'addPlot', 'addViewBox', 'addItem', 'getItem', 'addLayout', 'addLabel', 'removeItem', 'itemIndex', 'clear']:
            setattr(self, n, getattr(self.ci, n))
        self.setCentralItem(self.ci)

    def useOpenGL(self, b=True):
        """Overwrited to fix bad antialiasing while using openGL"""
        if b:
            if isOnWindows:
                v = QtOpenGL.QGLWidget()
            else:
                v = QtWidgets.QOpenGLWidget()
                fmt = QtGui.QSurfaceFormat()
                fmt.setSamples(8)
                fmt.setDepthBufferSize(24)
                fmt.setStencilBufferSize(8)  # Fixes antialias zoom error
                v.setFormat(fmt)
        else:
            v = QtGui.QWidget()
        self.setViewport(v)


class myGraphicsLayout(pg.GraphicsLayout):
    def __init__(self):
        super().__init__()
        self.layout.setHorizontalSpacing(-5)
        self.layout.setVerticalSpacing(-5)

    def addItem(self, item, row=None, col=None, rowspan=1, colspan=1):
        """
        Add an item to the layout and place it in the next available cell (or in the cell specified).
        The item must be an instance of a QGraphicsWidget subclass.
        """
        if row is None:
            row = self.currentRow
        if col is None:
            col = self.currentCol

        self.items[item] = []
        for i in range(rowspan):
            for j in range(colspan):
                row2 = row + i
                col2 = col + j
                if row2 not in self.rows:
                    self.rows[row2] = {}
                self.rows[row2][col2] = item
                self.items[item].append((row2, col2))

        self.layout.addItem(item, row, col, rowspan, colspan)
        self.nextRow()

    def set_graph_disposition(self, item, row=1, col=1, rowspan=1, colspan=1):
        """
        Function to modify the position of an item in the list

        :param item: WidgetPlotItem to set
        :param row: Row
        :param col: Column
        :param rowspan:
        :param colspan:
        :return:
        """
        # Remove from existing
        self.layout.removeItem(item)
        for r, c in self.items[item]:
            del self.rows[r][c]

        # Add new configuration
        for i in range(rowspan):
            for j in range(colspan):
                row2 = row + i
                col2 = col + j
                if row2 not in self.rows:
                    self.rows[row2] = {}
                self.rows[row2][col2] = item
                self.items[item].append((row2, col2))
        self.layout.addItem(item, row, col, rowspan, colspan)


class myItemSample(ItemSample):
    def __init__(self, item):
        super().__init__(item)
        self.offset = 13
        self.widthCell = 40

    def set_offset(self, offset):
        self.offset = offset

    def set_width_cell(self, width):
        self.widthCell = width

    # """Overwrites for text closer to the sample"""
    # def boundingRect(self):
    #     return QtCore.QRectF(0, 0, 0, 0)

    def paint(self, p, *args):
        """Overwrites to make matlab-like samples"""
        # p.setRenderHint(p.Antialiasing)  # only if the data is antialiased.
        opts = self.item.opts

        if opts.get('fillLevel', None) is not None and opts.get('fillBrush', None) is not None:
            p.setBrush(fn.mkBrush(opts['fillBrush']))
            p.setPen(fn.mkPen(None))
            p.drawPolygon(QtGui.QPolygonF([QtCore.QPointF(2, 18), QtCore.QPointF(18, 2), QtCore.QPointF(18, 18)]))

        p.save()
        linewidth = self.widthCell
        v_offset = self.offset

        opts = self.item.opts
        if not isinstance(self.item, ScatterPlotItem):
            pen = fn.mkPen(opts['pen'])
            pen.setWidthF(pen.widthF()*0.9)
            p.setPen(pen)
            p.drawLine(0, v_offset, linewidth, v_offset)

        p.restore()

        symbol = opts.get('symbol', None)
        if symbol is not None:
            if isinstance(symbol, list):
                symbol = symbol[0]

            if isinstance(self.item, PlotDataItem):
                opts = self.item.scatter.opts

            pen = fn.mkPen(opts['pen'])
            brush = fn.mkBrush(opts['brush'])
            size = opts['size'] * 1.0  # Scale factor for legend size

            p.translate(linewidth / 2, v_offset)
            drawSymbol(p, symbol, size, pen, brush)


class myLegend(pg.LegendItem):
    """Legend that fixes bugs (flush left + space) from pyqtgraph's legend"""
    def __init__(self, size=None, offset=(30, 30), is_light=False):
        super(myLegend, self).__init__(size, offset)
        self.layout.setVerticalSpacing(-8)

        self.is_light = is_light
        self.width_cell_sample = 25

    def set_space_sample_label(self, theSpace):
        """To set the gap between the sample and the label"""
        self.layout.setHorizontalSpacing(theSpace)

    def set_offset_sample(self, offset):
        """To tune the offset between the sample and the text"""
        for sample, label in self.items:
            sample.set_offset(offset)

    def set_width_cell_sample(self, width):
        """Set width of sample"""
        self.width_cell_sample = width
        self.apply_width_sample()

    def updateSize(self):
        if self.size is not None:
            return
        height = 0
        width = 0
        for sample, label in self.items:
            height += max(sample.boundingRect().height(), label.height())
            width = max(width, sample.boundingRect().width() + label.width())  # + 13
        self.setGeometry(0, 0, width, height)

    def addItem(self, item, name):
        """Overwrites to flush left"""
        # label = LabelItem(name)
        label = myLabelItem(name, justify='left')
        width_cell = self.width_cell_sample
        if isinstance(item, ItemSample):
            sample = item
        else:
            sample = myItemSample(item)
            sample.set_width_cell(width_cell*1)
        row = self.layout.rowCount()

        self.items.append((sample, label))
        self.layout.addItem(sample, row, 0)
        self.layout.setColumnFixedWidth(0, width_cell)
        self.layout.setColumnFixedWidth(1, 10)
        self.layout.setHorizontalSpacing(0.0)  # No spacing between sample and text
        self.layout.addItem(label, row, 1)
        self.layout.setRowFixedHeight(row, 15)

        self.updateSize()

    def apply_width_sample(self):
        width_cell = self.width_cell_sample
        for sample, label in self.items:
            sample.set_width_cell(width_cell*1)
        self.layout.setColumnFixedWidth(0, width_cell)
        self.updateSize()

    def set_font(self, font_size, font_color, fontname=None):
        for item in self.items:
            for single_item in item:
                if isinstance(single_item, pg.LabelItem):
                    if fontname is not None:
                        single_item.setText(single_item.text, **{'color': font_color, 'size': str(font_size) + 'pt', 'font-family': fontname})
                    else:
                        single_item.setText(single_item.text, **{'color': font_color, 'size': str(font_size) + 'pt'})

    def paint(self, p, *args):
        """Overwrited to select background color"""
        if len(self.items):
            if self.is_light:
                p.setPen(pg.mkPen(100, 100, 100, 255))
                p.setBrush(pg.mkBrush(255, 255, 255, 255))
            else:
                p.setPen(pg.mkPen(100, 100, 100, 255))
                p.setBrush(pg.mkBrush(10, 10, 10, 255))
            offset_x = 0
            offset_y = 20
            boundingRect = QtCore.QRectF(0, offset_y/2, self.width()-offset_x, self.height()-offset_y)
            p.drawRect(boundingRect)  # self.boundingRect())

    def set_position(self, position, offset):
        """
        Set the position of the legend, in a corner.

        :param position: String (NW, NE, SW, SE), indicates which corner the legend is close
        :param offset: Tuple (xoff, yoff), x and y offset from the edge
        :return:
        """
        offPosMult = 1

        if position == "NW":
            quartil_parent = (0, 0)
            oxMult, oyMult = 1, 1
        elif position == "NE":
            quartil_parent = (1, 0)
            oxMult, oyMult = -1, 1
        elif position == "SW":
            quartil_parent = (0, 1)
            oxMult, oyMult = 1, -1
            offPosMult = -1
        elif position == "SE":
            quartil_parent = (1, 1)
            oxMult, oyMult = -1, -1
            offPosMult = -1
        else:
            print("Could not place legend : position unknown (please use NW, NE, SW or SE)")
            return
        self.anchor(quartil_parent, quartil_parent, offset=(offset[0] * oxMult, offPosMult * (-10) + offset[1] * oyMult))  # * -10: corresponds to offset_y defined in paint


class myLabelItem(pg.LabelItem):
    def setText(self, text, **args):
        """ Overwrited to add font-family to options """
        self.text = text
        opts = self.opts
        for k in args:
            opts[k] = args[k]

        optlist = []

        color = self.opts['color']
        if color is None:
            color = pg.getConfigOption('foreground')
        color = fn.mkColor(color)
        optlist.append('color: #' + fn.colorStr(color)[:6])
        if 'size' in opts:
            optlist.append('font-size: ' + opts['size'])
        if 'bold' in opts and opts['bold'] in [True, False]:
            optlist.append('font-weight: ' + {True: 'bold', False: 'normal'}[opts['bold']])
        if 'italic' in opts and opts['italic'] in [True, False]:
            optlist.append('font-style: ' + {True: 'italic', False: 'normal'}[opts['italic']])
        if 'font-family' in opts:
            optlist.append('font-family: ' + opts['font-family'])
        full = "<span style='%s'>%s</span>" % ('; '.join(optlist), text)
        # print full
        self.item.setHtml(full)
        self.updateMin()
        self.resizeEvent(None)
        self.updateGeometry()


class myAxis(pg.AxisItem):
    update_label = QtCore.pyqtSignal(str)

    def __init__(self, orientation):
        super().__init__(orientation)
        self.nudges = {'left': (-10, 0), 'right': (5, 0), 'bottom': (0, -5), 'top': (0, 5)}
        self.update_label.connect(self.label.setHtml)
        # self.style.update({'textFillLimits': [(0, 0.8),    ## never fill more than 80% of the axis
        #         (4, 0.4),    ## If we already have 4 ticks with text, fill no more than 40% of the axis
        #         (6, 0.2)]})

    def _updateLabel(self):
        # self.label.setHtml(self.labelString())
        self.update_label.emit(self.labelString())  # Use of a signal/slot instead of directly changing the value
        self._adjustSize()
        self.picture = None
        self.update()

    def get_label_pos(self):
        """Overwrited to place label closer to the axis"""
        br = self.label.boundingRect()
        x, y = 0, 0
        if self.orientation == 'left':
            y, x = int(self.size().height()/2 + br.width()/2), 0
        elif self.orientation == 'right':
            y, x = int(self.size().height()/2 + br.width()/2), int(self.size().width()-br.height())
        elif self.orientation == 'top':
            y, x = 0, int(self.size().width()/2. - br.width()/2.)
        elif self.orientation == 'bottom':
            x, y = int(self.size().width()/2. - br.width()/2.), int(self.size().height()-br.height())
        try:
            return int(x + self.nudges[self.orientation][0]), int(y - self.nudges[self.orientation][1])
        except AttributeError:
            return x, y

    def resizeEvent(self, ev=None):
        """Overwrited to place label closer to the axis"""
        p = QtCore.QPointF(0, 0)
        x_pos, y_pos = self.get_label_pos()
        p.setX(int(x_pos))
        p.setY(int(y_pos))
        self.label.setPos(p)
        self.picture = None

    def set_label_pos(self, orientation, x_offset=0, y_offset=0):
        self.nudges[orientation] = (x_offset, y_offset)

    def set_number_ticks(self, number):
        self.style.update({'textFillLimits': [(number, 0)]})



import PyQt5.QtCore

PyQt5.QtCore.QCoreApplication.setAttribute(PyQt5.QtCore.Qt.AA_ShareOpenGLContexts)

from .displayCollections import CollectionDisplayer
from .displaySensitivity import SensitivityDisplayer, analyse_sobol_plot_indices, analyse_sobol_plot_convergence, analyse_sobol_plot_2ndOrder_indices
from .displayOptimization import OptimizationDisplayer
from .viewOptimizationResults import ViewOptimizationResults
from .fastPlot import *
from .fastPlot3 import *

# Submodules import
from .graphs.widget_graphsVisual import *
from .widgets import *
from .onclick import *
from .selector import *





from PyQt5.QtWidgets import QSlider
from PyQt5.QtCore import pyqtSignal


class widget_doubleSlider(QSlider):
    doubleValueChanged = pyqtSignal(float)

    def __init__(self, decimals=3, *args, **kwargs):
        """
        Like a QSlider but accepts float numbers
        :param decimals: number of decimals to account for
        :param args: args to pass to QSlider
        :param kwargs: kwargs to pass to QSlider
        """
        super().__init__(*args, **kwargs)
        self._multi = 10 ** decimals

        self.valueChanged.connect(self.emitDoubleValueChanged)

    def emitDoubleValueChanged(self):
        value = float(super().value())/self._multi
        self.doubleValueChanged.emit(value)

    def value(self):
        return float(super().value()) / self._multi

    def setMinimum(self, value):
        return super().setMinimum(value * self._multi)

    def setMaximum(self, value):
        return super().setMaximum(value * self._multi)

    def setSingleStep(self, value):
        return super().setSingleStep(value * self._multi)

    def singleStep(self):
        return float(super().singleStep()) / self._multi

    def setValue(self, value):
        super().setValue(int(value * self._multi))


from abc import ABCMeta, abstractmethod

class OnclickInterface(metaclass=ABCMeta):
    """Interface class for the action to perform when a point is clicked"""

    @abstractmethod
    def graph_clicked(self, theGraphsVisual, index_graph, index_trace, indices_points):
        """
        Action to perform when a graph is clicked

        :param theGraphsVisual: class widget_graphs_visual that has called the method
        :param index_graph: Index of the graph that has been clicked
        :param index_trace: Index of the trace that has been clicked
        :param indices_points: graph Indices of the points that have been clicked
        :return:
        """
        pass

    @abstractmethod
    def get_name(self):
        pass


from optimeed.visualize.widgets import Widget_text_scrollable, Widget_lineDrawer
from .animationGUI import AnimationGUI


class Animate_openGL(AnimationGUI):
    """Implements :class:`~DataAnimationVisuals` to show opengl drawing"""
    def __init__(self, theOpenGLWidget, theId=0, window_title='Animation'):
        super().__init__(theId, window_title)
        self.theOpenGLWidget = theOpenGLWidget
        self.horizontal_layout_visu.addWidget(self.theOpenGLWidget)

    def update_widget_w_animation(self, key, index, the_data_animation):
        self.theOpenGLWidget.set_deviceToDraw(the_data_animation.get_element_animations(0, index))

    def export_widget(self, painter):
        pass

    def delete_key_widgets(self, key):
        pass


class Animate_openGL_and_text(Animate_openGL):
    """Implements :class:`~DataAnimationVisuals` to show opengl drawing and text"""
    def __init__(self, *args, is_light=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.textWidget = Widget_text_scrollable('', is_light=is_light, convertToHtml=False)
        self.horizontal_layout_visu.addWidget(self.textWidget)

    def update_widget_w_animation(self, key, index, the_data_animation):
        super().update_widget_w_animation(key, index, the_data_animation)
        self.textWidget.set_text(the_data_animation.get_element_animations(1, index), convertToHtml=True)

    def get_interesting_elements(self, devices_list):
        all_listOfText = [str(device) for device in devices_list]
        return [devices_list, all_listOfText]


class Animate_lines(AnimationGUI):
    """Implements :class:`~DataAnimationVisuals` to show drawing made out of lines (:class:`~optimeed.visualize.gui.widgets.widget_line_drawer`)"""
    def __init__(self, get_lines_method, is_light=True, theId=0, window_title='Animation'):
        """

        :param get_lines_method: Method that takes Device as argument and returns list of lines
        :param is_light:
        :param theId:
        :param window_title:
        """
        super().__init__(theId, window_title)
        self.qtLinesDrawer = Widget_lineDrawer(is_light=is_light)
        self.horizontal_layout_visu.addWidget(self.qtLinesDrawer)
        self.get_lines_method = get_lines_method

    def export_widget(self, painter):
        self.qtLinesDrawer.paintEvent(event=None, painter=painter)

    def delete_key_widgets(self, key):
        self.qtLinesDrawer.delete_lines(key)

    def update_widget_w_animation(self, key, index, the_data_animation):
        self.qtLinesDrawer.set_lines(the_data_animation.get_element_animations(0, index), key_id=key, pen=the_data_animation.get_base_pen())

    def get_interesting_elements(self, devices_list):
        all_listsOfLines = [self.get_lines_method(device) for device in devices_list]
        all_listOfText = [str(device) for device in devices_list]
        return [all_listsOfLines, all_listOfText]


class Animate_lines_and_text(Animate_lines):
    """Same as :class:`~DataAnimationLines` but also with text"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.textWidget = Widget_text_scrollable('', is_light=kwargs.get("is_light", True), convertToHtml=False)
        self.horizontal_layout_visu.addWidget(self.textWidget)

    def update_widget_w_animation(self, key, index, the_data_animation):
        self.qtLinesDrawer.set_lines(the_data_animation.get_element_animations(0, index), key_id=key, pen=the_data_animation.get_base_pen())
        self.textWidget.set_text(the_data_animation.get_element_animations(1, index), convertToHtml=True)




from .traceVisual import TraceVisual
from .pyqtgraphRedefine import myLegend, myAxis
from optimeed.core.tools import printIfShown, SHOW_WARNING
from PyQt5 import QtGui
from optimeed.core.color_palette import default_palette


class GraphVisual:
    """Provide an interface to a graph. A graph contains traces."""
    def __init__(self, theWidgetGraphVisual):
        """

        :param theWidgetGraphVisual: :class:`~optimeed.visualize.gui.widgets.widget_graphs_visual.widget_graphs_visual`
        """
        self.theWidgetGraphVisual = theWidgetGraphVisual

        self.traceVisuals = dict()
        self.theWGPlot = self.theWidgetGraphVisual.canvasWidget.addPlot(axisItems={'bottom': myAxis(orientation='bottom'), 'left': myAxis(orientation='left')})  # Class PLOT ITEM
        # self.theWGPlot.disableAutoRange()
        self.theWGPlot.legend = myLegend(is_light=self.theWidgetGraphVisual.is_light)
        self.theWGPlot.legend.setParentItem(self.theWGPlot.vb)

        self.theWGPlot.showGrid(x=True, y=True)
        self.set_fontTicks(11)
        self.palette = default_palette

        # Set font label
        if self.theWidgetGraphVisual.is_light:
            label_color = '#000'
        else:
            label_color = '#FFF'
        self.labelStyle = None
        self.set_fontLabel(12, label_color)

    def set_fontTicks(self, fontSize, fontname=None):
        """ Set font of the ticks

        :param fontSize: Size of the font
        :param fontname: Name of the font
        """
        font = QtGui.QFont()
        font.setPointSizeF(fontSize)
        if fontname is not None:
            font.setFamily(fontname)
        self.theWGPlot.getAxis("bottom").tickFont = font
        self.theWGPlot.getAxis("left").tickFont = font

    def set_numberTicks(self, number, axis):
        """ Set the number of ticks to be displayed

        :param number: Number of ticks for the axis
        :param axis: Axis (string, "bottom", "left", "right", "top")
        :return:
        """
        self.theWGPlot.getAxis(axis).set_number_ticks(number)

    def set_fontLabel(self, fontSize, color='#000', fontname=None):
        """Set font of the axis labels

        :param fontSize: font size
        :param color: color in hexadecimal (str)
        :param fontname: name of the font"""
        self.labelStyle = {'font-size': str(fontSize) + 'pt', 'color': color, "font-family": fontname}

    def get_legend(self) -> myLegend:
        """ Get the legend"""
        return self.theWGPlot.legend

    def get_axis(self, axis) -> myAxis:
        """ Get the axis

        :param axis: Axis (string, "bottom", "left", "right", "top")
        :return: axis object
        """
        return self.theWGPlot.getAxis(axis)

    def set_fontLegend(self, font_size, font_color, fontname=None):
        self.get_legend().set_font(font_size, font_color, fontname)

    def set_label_pos(self, orientation, x_offset=0, y_offset=0):
        self.theWGPlot.getAxis(orientation).set_label_pos(orientation, x_offset, y_offset)

    def set_color_palette(self, palette):
        self.palette = palette
        self.apply_palette()

    def apply_palette(self):
        numb_color_to_set = len(self.get_all_traces())
        theColors = self.palette(numb_color_to_set)
        for i, trace in enumerate(self.get_all_traces().values()):
            if trace.get_color() is None:
                trace.set_color(theColors[i])

    def hide_axes(self):
        self.theWGPlot.showGrid(x=False, y=False)
        self.theWGPlot.getAxis("bottom").hide()
        self.theWGPlot.getAxis("left").hide()

    def add_feature(self, theFeature):
        """To add any pyqtgraph item to the graph"""
        self.theWGPlot.addItem(theFeature)

    def remove_feature(self, theFeature):
        """To remove any pyqtgraph item from the graph"""
        self.theWGPlot.removeItem(theFeature)

    def add_data(self, idGraph, theData):
        theTrace = TraceVisual(theData, self.theWGPlot, highlight_last=self.theWidgetGraphVisual.highlight_last)
        self.add_trace(idGraph, theTrace)
        if len(self.traceVisuals) <= 1:
            self.set_graph_properties(theTrace)

    def set_graph_properties(self, theTrace):
        """This function is automatically called on creation of the graph"""
        self.theWGPlot.setLabel('bottom', text=theTrace.theData.get_x_label(), **self.labelStyle)
        self.theWGPlot.setLabel('left', text=theTrace.theData.get_y_label(), **self.labelStyle)

        self.set_lims(theTrace.theData.get_xlim(), theTrace.theData.get_ylim())

    def set_lims(self, xlim, ylim):
        """Set limits of the graphs, xlim or ylim = [val_low, val_high]. Or None."""
        # if xlim is not None:
        #     self.theWGPlot.setLimits(xMin=xlim[0], xMax=xlim[1])
        # if ylim is not None:
        #     self.theWGPlot.setLimits(yMin=ylim[0], yMax=ylim[1])
        #
        if xlim is not None or ylim is not None:
            self.theWGPlot.setRange(xRange=xlim, yRange=ylim)
        # self.theWGPlot.setAutoVisible(x=xlim is None, y=ylim is None)

    def add_trace(self, idTrace, theTrace):
        """Add a :class:`~optimeed.visualize.gui.widgets.graphsVisualWidget.TraceVisual.TraceVisual` to the graph, with index idTrace"""
        self.traceVisuals[idTrace] = theTrace
        theTrace.thePlotItem.sigPointsClicked.connect(self.theWidgetGraphVisual.on_click)  # Connect the function "on click"

    def set_legend(self):
        """Set default legend options (color and font)"""
        if self.theWidgetGraphVisual.is_light:
            legend_color = '#111'
        else:
            legend_color = '#DDD'

        self.set_fontLegend(10, legend_color)

    def set_title(self, titleName, **kwargs):
        """Set title of the graph

        :param titleName: title to set
        """
        self.theWGPlot.setTitle(title=titleName, **kwargs)

    def get_trace(self, idTrace) -> TraceVisual:
        """Return the :class:`~optimeed.visualize.gui.widgets.graphsVisualWidget.TraceVisual.TraceVisual` correspondong to the index idTrace"""
        return self.traceVisuals[idTrace]

    def get_all_traces(self):
        """Return a dictionary {idtrace: :class:`~optimeed.visualize.gui.widgets.graphsVisualWidget.TraceVisual.TraceVisual`}."""
        return self.traceVisuals

    def delete_trace(self, idTrace):
        """Delete the trace of index idTrace"""
        try:
            self.theWGPlot.removeItem(self.traceVisuals[idTrace].thePlotItem)
            self.traceVisuals.pop(idTrace)
            # self.signal_must_update.emit()
        except KeyError:
            printIfShown("Trace already deleted", SHOW_WARNING)

    def delete(self):
        """Delete the graph"""
        self.theWGPlot.deleteLater()
        self.theWidgetGraphVisual.canvasWidget.centralWidget.layout.removeItem(self.theWGPlot)

    def linkXToGraph(self, graph):
        """Link the axis of the current graph to an other :class:`GraphVisual`"""
        self.theWGPlot.setXLink(graph.theWGPlot)

    def update(self):
        """Update the traces contained in the graph"""
        for trace in self.get_all_traces().values():
            self.set_graph_properties(trace)
            trace.signal_must_update.emit()

    def fast_update(self):
        """Same as :meth:`~GraphVisual.update` but faster. This is NOT thread safe (cannot be called a second time before finishing operation)"""
        for trace in self.get_all_traces().values():
            trace.updateTrace()

    def axis_equal(self):
        self.theWGPlot.vb.setAspectLocked(True)

    def log_mode(self, x=False, y=False):
        self.theWGPlot.setLogMode(x, y)

    def grid_off(self):
        """Turn off grid"""
        self.theWGPlot.showGrid(x=False, y=False)



from optimeed.visualize.widgets.openGL.contextHandler import ContextHandler, SpecialButtonsMapping
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QOpenGLWidget
from PyQt5.QtGui import QSurfaceFormat
from PyQt5.Qt import QSize


class Widget_openGL(QOpenGLWidget):
    """Interface that provides opengl capabilities.
    Ensures zoom, light, rotation, etc."""
    def __init__(self, parent=None):
        super().__init__(parent)
        self.theContextHandler = ContextHandler()
        self.theContextHandler.set_specialButtonsMapping(self._get_specialButtonsMapping())
        self.setFocusPolicy(Qt.StrongFocus)
        theFormat = QSurfaceFormat()
        theFormat.setSamples(10)
        self.setFormat(theFormat)
        self.resize(self.sizeHint())

    def sizeHint(self):
        return QSize(400, 400)

    def minimumSizeHint(self):
        return QSize(400, 400)

    def set_deviceDrawer(self, theDeviceDrawer):
        """Set a drawer :class:`optimeed.visualize.widgets.openGL.deviceDrawerInterface.DeviceDrawerInterface`"""
        self.theContextHandler.set_deviceDrawer(theDeviceDrawer)

    def set_deviceToDraw(self, theDeviceToDraw):
        """Set the device to draw"""
        self.theContextHandler.set_deviceToDraw(theDeviceToDraw)
        self.update()

    @staticmethod
    def _get_specialButtonsMapping():
        theMapping = SpecialButtonsMapping()
        theMapping.KEY_UP = Qt.Key_Up
        theMapping.KEY_DOWN = Qt.Key_Down
        theMapping.KEY_LEFT = Qt.Key_Left
        theMapping.KEY_RIGHT = Qt.Key_Right
        theMapping.KEY_TAB = Qt.Key_Tab
        theMapping.MOUSE_LEFT = Qt.LeftButton
        theMapping.MOUSE_RIGHT = Qt.RightButton
        return theMapping

    def initializeGL(self):
        self.theContextHandler.initialize()
        self.update()

    def paintGL(self):
        self.theContextHandler.redraw()
        # # Uncomment to print text
        # painter = QPainter(self)
        # for text in self.theContextHandler.get_text_to_write():
        #     fontColor = QColor(*text.color)
        #     painter.setPen(fontColor)
        #     painter.drawText(text.position[0], text.position[1], text.theStr)
        # painter.end()

    def resizeGL(self, w, h):
        self.theContextHandler.resizeWindowAction(w, h)
        self.update()

    def mousePressEvent(self, event):
        self.theContextHandler.mouseClicAction(event.button(), event.x(), event.y())
        self.update()

    def mouseMoveEvent(self, event):
        self.theContextHandler.mouseMotionAction(event.x(), event.y())
        self.update()

    def keyPressEvent(self, event):
        self.theContextHandler.keyboardPushAction(event.key())
        super().keyPressEvent(event)
        self.update()

    def wheelEvent(self, QWheelEvent):
        self.theContextHandler.mouseWheelAction(float(QWheelEvent.angleDelta().y())/80.0)
        self.update()



import nlopt
import numpy as np

from optimeed.core.tools import isNonePrintMessage, printIfShown, SHOW_ERROR, SHOW_WARNING, indentParagraph
from .algorithmInterface import AlgorithmInterface
from optimeed.optimize.optiVariable import Binary_OptimizationVariable, Integer_OptimizationVariable, Real_OptimizationVariable
from .convergence import EvolutionaryConvergence
from optimeed.core import Option_class, Option_int, Option_str


class ConvergenceManager:
    def __init__(self):
        self.objectives = []
        self.popSize = 0
        self.convergence = EvolutionaryConvergence()

    def add_point(self, newObj):
        self.objectives.append([newObj])
        if len(self.objectives) == self.popSize:
            self.convergence.set_curr_step(self.objectives, [[]]*len(self.objectives))
            self.objectives = []

    def set_pop_size(self, popSize):
        self.popSize = popSize


class NLOpt_Algorithm(AlgorithmInterface, Option_class):
    ALGORITHM = 0
    POPULATION_SIZE = 1

    def __init__(self):
        super().__init__()
        self.maxTime = None  # set by set_maxtime
        self.evaluationFunction = None  # set by set_evaluationFunction
        self.theDimensionalConstraints = None

        self.add_option(self.ALGORITHM, Option_str("Algorithm. Recommended local/global: LN_SBPLX/GN_CRS2_LM . Consult NLopt help page for more.", 'GN_CRS2_LM'))
        self.add_option(self.POPULATION_SIZE, Option_int("Size of the population (stochastic algorithms only)", 20))

        self.convManager = ConvergenceManager()
        self.curr_pop = 0

        self.initialVectorGuess = None
        self.listOfOptimizationVariables = None

    def initialize(self, initialVectorGuess, listOfOptimizationVariables):
        self.initialVectorGuess = initialVectorGuess
        self.listOfOptimizationVariables = listOfOptimizationVariables

    def compute(self):
        # Check if variables properly initialized, otherwise return
        notInit_message = " not initialized in NLOpt_Algorithm. Aborting optimization loop."
        if isNonePrintMessage(self.maxTime, "maxTime" + notInit_message, SHOW_ERROR):
            return
        if isNonePrintMessage(self.evaluationFunction, "evaluationFunction" + notInit_message, SHOW_ERROR):
            return
        if self.get_option_value(self.POPULATION_SIZE) <= len(self.initialVectorGuess):
            raise ValueError("Population size must be bigger than the number of parameters of this algorithm. Use options")

        self.convManager.set_pop_size(self.get_option_value(self.POPULATION_SIZE))
        # If everything exists: create nlopt and launch it
        numberOfOptimizationParameters = len(self.initialVectorGuess)
        theOptimizationAlgorithm = nlopt.opt(getattr(nlopt, self.get_option_value(self.ALGORITHM)), numberOfOptimizationParameters)

        theOptimizationAlgorithm.set_min_objective(self.evaluationFunction)
        theOptimizationAlgorithm.set_maxtime(float(self.maxTime))

        self.__set_optimizationVariables(self.listOfOptimizationVariables, theOptimizationAlgorithm)

        theOptimizationAlgorithm.set_population(int(self.get_option_value(self.POPULATION_SIZE)))
        return [theOptimizationAlgorithm.optimize(np.array(self.initialVectorGuess).astype(float))]

    @staticmethod
    def __set_optimizationVariables(listOfOptimizationVariables, theOptimizationAlgorithm):
        lowerBounds = []
        upperBounds = []
        for optimizationVariable in listOfOptimizationVariables:
            if type(optimizationVariable) is Real_OptimizationVariable:
                lowerBounds.append(optimizationVariable.get_min_value())
                upperBounds.append(optimizationVariable.get_max_value())
            elif type(optimizationVariable) is Integer_OptimizationVariable:
                printIfShown("Integer variables not natively supported by the optimization algorithm", SHOW_WARNING)
                lowerBounds.append(optimizationVariable.get_min_value())
                upperBounds.append(optimizationVariable.get_max_value())
            elif type(optimizationVariable) is Binary_OptimizationVariable:
                printIfShown("Boolean variables not natively supported by the optimization algorithm", SHOW_WARNING)
                lowerBounds.append(0)
                upperBounds.append(1)
            else:
                raise ValueError("Optimization algorithm not working with this kind of optimization variables")

        theOptimizationAlgorithm.set_lower_bounds(np.array(lowerBounds).astype(float))
        theOptimizationAlgorithm.set_upper_bounds(np.array(upperBounds).astype(float))

    def set_evaluationFunction(self, evaluationFunction, callback_on_evaluate, numberOfObjectives, _numberOfConstraints):
        if numberOfObjectives > 1:
            printIfShown("Optimization algorithm does not support true multiobjective. Using sum instead", SHOW_WARNING)

        def adapted_evaluationFunction(x_in, _grad):
            returnedValues = evaluationFunction(x_in)

            objs = returnedValues["objectives"]
            const = returnedValues["constraints"]
            sum_const = 0
            for constraint in const:
                if constraint > 0:
                    sum_const += constraint*7000

            callback_on_evaluate(returnedValues)
            objective = np.float(sum(objs) + sum_const)

            self.convManager.add_point(objective)
            return objective
        self.evaluationFunction = adapted_evaluationFunction

    def set_maxtime(self, maxTime):
        self.maxTime = maxTime

    def __str__(self):
        theStr = ''
        theStr += "NLOpt optimization library\n"
        theStr += indentParagraph(super().__str__(), indent_level=1)
        return theStr

    def get_convergence(self):
        return self.convManager.convergence



from .pyqtgraph import ColorMap
import collections
from optimeed.core import printIfShown, SHOW_INFO
try:
    from matplotlib import cm
    has_matplotlib = True
except ImportError:
    has_matplotlib = False
    printIfShown("Matplotlib library not found -> specific 3D colormap has been deactivated", SHOW_INFO)
import numpy as np

try:
    sequence = collections.Sequence
except AttributeError:
    sequence = collections.abc.Sequence


def matplotlib_colormap_to_pg_colormap(colormap_name, n_ticks=16):
    pos, rgba_colors = zip(*cmapToColormap(getattr(cm, colormap_name), n_ticks))
    pgColormap = ColorMap(pos, rgba_colors)
    return pgColormap


def cmapToColormap(cmap, nTicks=16):
    """
    Converts a Matplotlib cmap to pyqtgraphs colormaps. No dependency on
    matplotlib.
    Parameters:
        *cmap*: Cmap object. Imported from matplotlib.cm.*
        *nTicks*: Number of ticks to create when dict of functions is used.
        Otherwise unused.
    author: Sebastian Hoefer
    """
    rgb_list = list()
    # Case #1: a dictionary with 'red'/'green'/'blue' values as list of ranges (e.g. 'jet')
    # The parameter 'cmap' is a 'matplotlib.colors.LinearSegmentedColormap' instance ...
    if hasattr(cmap, '_segmentdata'):
        colordata = getattr(cmap, '_segmentdata')
        if ('red' in colordata) and isinstance(colordata['red'], sequence):

            # collect the color ranges from all channels into one dict to get unique indices
            posDict = {}
            for idx, channel in enumerate(('red', 'green', 'blue')):
                for colorRange in colordata[channel]:
                    posDict.setdefault(colorRange[0], [-1, -1, -1])[idx] = colorRange[2]

            indexList = list(posDict.keys())
            indexList.sort()
            # interpolate missing values (== -1)
            for channel in range(3):  # R,G,B
                startIdx = indexList[0]
                emptyIdx = []
                for curIdx in indexList:
                    if posDict[curIdx][channel] == -1:
                        emptyIdx.append(curIdx)
                    elif curIdx != indexList[0]:
                        for eIdx in emptyIdx:
                            rPos = (eIdx - startIdx) / (curIdx - startIdx)
                            vStart = posDict[startIdx][channel]
                            vRange = (posDict[curIdx][channel] - posDict[startIdx][channel])
                            posDict[eIdx][channel] = rPos * vRange + vStart
                        startIdx = curIdx
                        del emptyIdx[:]
            for channel in range(3):  # R,G,B
                for curIdx in indexList:
                    posDict[curIdx][channel] *= 255

            rgb_list = [[i, posDict[i]] for i in indexList]

        # Case #2: a dictionary with 'red'/'green'/'blue' values as functions (e.g. 'gnuplot')
        elif ('red' in colordata) and isinstance(colordata['red'], collections.Callable):
            indices = np.linspace(0., 1., nTicks)
            luts = [np.clip(np.array(colordata[rgb](indices), dtype=np.float), 0, 1) * 255 for rgb in ('red', 'green', 'blue')]
            rgb_list = zip(indices, list(zip(*luts)))

    # If the parameter 'cmap' is a 'matplotlib.colors.ListedColormap' instance, with the attributes 'colors' and 'N'
    elif hasattr(cmap, 'colors') and hasattr(cmap, 'N'):
        colordata = getattr(cmap, 'colors')
        # Case #3: a list with RGB values (e.g. 'seismic')
        if len(colordata[0]) == 3:
            indices = np.linspace(0., 1., len(colordata))
            scaledRgbTuples = [(rgbTuple[0] * 255, rgbTuple[1] * 255, rgbTuple[2] * 255) for rgbTuple in colordata]
            rgb_list = zip(indices, scaledRgbTuples)

        # Case #4: a list of tuples with positions and RGB-values (e.g. 'terrain')
        # -> this section is probably not needed anymore!?
        elif len(colordata[0]) == 2:
            rgb_list = [(idx, (vals[0] * 255, vals[1] * 255, vals[2] * 255)) for idx, vals in colordata]

    # Case #X: unknown format or datatype was the wrong object type
    else:
        raise ValueError("[cmapToColormap] Unknown cmap format or not a cmap!")

    # Convert the RGB float values to RGBA integer values
    return list([(pos, (int(r), int(g), int(b), 255)) for pos, (r, g, b) in rgb_list])



from .onclickInterface import OnclickInterface


class Onclick_changeSymbol(OnclickInterface):
    """On Click: Change the symbol of the point that is clicked"""

    def __init__(self, theLinkDataGraph):
        """

        :param theLinkDataGraph: :class:`~optimeed.visualize.high_level.LinkDataGraph.LinkDataGraph`
        """
        self.theLinkDataGraph = theLinkDataGraph

    def graph_clicked(self, theGraphVisual, index_graph, index_trace, indices_points):
        theTrace = theGraphVisual.get_graph(index_graph).get_trace(index_trace)
        for index_point in indices_points:
            theTrace.set_symbol(index_point, 'x')

    def get_name(self):
        return "Change symbol"



from abc import ABCMeta, abstractmethod


class OnselectInterface(metaclass=ABCMeta):
    @abstractmethod
    def get_name(self):
        """Get the name of the action

        :return: string
        """
        pass

    @abstractmethod
    def selector_updated(self, selection_name, the_collection, selected_indices, not_selected_indices):
        """
        Action to perform once the data have been selected
        :param selection_name: name of the selection (deprecated ?)
        :param the_collection: the collection
        :param selected_indices: indices of selected data
        :param not_selected_indices: indices of data NOT selected
        :return: identifier that can later be used with cancel_selector
        """
        pass

    @abstractmethod
    def cancel_selector(self, selection_identifier):
        """
        Action to perform when data stopped being selected
        :param selection_identifier: identifier that was returned by selector_updated
        :return:
        """
        pass



import math

from OpenGL.GL import *
from OpenGL.GLU import *

from .triangulate_polygon import meshPolygon


def draw_closedPolygon(xClockWise, yClockWise):
    theTriList = meshPolygon(xClockWise, yClockWise)

    glBegin(GL_TRIANGLES)
    for (p1, p2, p3) in theTriList:
        glVertex2f(p1[0], p1[1])
        glVertex2f(p2[0], p2[1])
        glVertex2f(p3[0], p3[1])
    glEnd()


def draw_extrudeZ(xList, yList, zExtrude):
    glBegin(GL_QUAD_STRIP)
    for i in range(len(xList)):
        glVertex3f(xList[i], yList[i], 0.0)
        glVertex3f(xList[i], yList[i], zExtrude)
    glEnd()


def draw_triList(theTriList):
    glBegin(GL_TRIANGLES)
    for (p1, p2, p3) in theTriList:
        glVertex2f(p1[0], p1[1])
        glVertex2f(p2[0], p2[1])
        glVertex2f(p3[0], p3[1])
    glEnd()


def draw_lines(x, z):
    if len(x) < 2:
        return
    glBegin(GL_LINE_STRIP)
    for i in range(len(x)):
        glVertex2f(x[i], z[i])
    glEnd()


def draw_spiralSheet(innerRadius, thickness, length, theAngle, n, reverseDirection=False):
    if reverseDirection:
        glFrontFace(GL_CW)

    a = innerRadius
    b = thickness / (2 * math.pi)
    glBegin(GL_TRIANGLE_STRIP)
    for i in range(n + 1):
        angle = i / n * theAngle
        radius = a + b * angle
        x = radius * math.cos(angle)
        y = radius * math.sin(angle)
        glVertex(x, y, length)
        glVertex(x, y, 0)
        glNormal3f(x, y, 0)
    glEnd()

    glFrontFace(GL_CCW)


def draw_spiralFront(innerRadius, thicknessMaterial, thicknessSpiral, z0, theAngle, n, reverseDirection=False):
    a = innerRadius
    b = thicknessSpiral / (2 * math.pi)
    glBegin(GL_QUAD_STRIP)
    numberOfPoints = n
    rIn = innerRadius
    for i in range(numberOfPoints + 1):
        if reverseDirection:
            angle = (n - i) / n * theAngle
        else:
            angle = i / n * theAngle
        rIn = a + b * angle
        rOut = rIn + thicknessMaterial
        glVertex3f(rIn * math.cos(angle), rIn * math.sin(angle), z0)
        glVertex3f(rOut * math.cos(angle), rOut * math.sin(angle), z0)
        glNormal3f(0, 0, 1)
    glEnd()


def draw_spiralFull(innerRadius, outerRadius, thicknessMaterial, thicknessSpiral, length, n):
    if outerRadius - innerRadius < thicknessMaterial:
        return
    nTurns = ((outerRadius - thicknessMaterial) - innerRadius) / thicknessSpiral
    theta_tot = nTurns * 2 * math.pi
    draw_spiralSheet(innerRadius, thicknessSpiral, length, theta_tot, n)
    draw_spiralSheet(innerRadius + thicknessMaterial, thicknessSpiral, length, theta_tot, n)
    draw_spiralFront(innerRadius, thicknessMaterial, thicknessSpiral, 0, theta_tot, n)
    draw_spiralFront(innerRadius, thicknessMaterial, thicknessSpiral, length, theta_tot, n)


def draw_spiral(innerRadius, outerRadius, thicknessMaterial, thicknessSpiral, length, cutAngle, n):
    if outerRadius - innerRadius < thicknessMaterial:
        return

    glPushMatrix()
    glTranslate(0, 0, -length / 2)

    a = innerRadius
    b = thicknessSpiral / (2 * math.pi)

    nTurns = ((outerRadius - thicknessMaterial) - innerRadius) / thicknessSpiral
    theta_last = (nTurns - math.floor(nTurns)) * 2 * math.pi
    for i in range(int(math.floor(nTurns))):
        currentAngle = i * 2 * math.pi
        rIn = a + b * currentAngle
        currentAngle2 = i * 2 * math.pi + cutAngle
        rIn2 = a + b * currentAngle2

        draw_spiralSheet(rIn, thicknessSpiral, length, cutAngle, n, False)
        draw_spiralSheet(rIn + thicknessMaterial, thicknessSpiral, length, cutAngle, n, True)
        draw_spiralFront(rIn, thicknessMaterial, thicknessSpiral, 0, cutAngle, n, True)
        draw_spiralFront(rIn, thicknessMaterial, thicknessSpiral, length, cutAngle, n)
        draw_rectangle(rIn, length, thicknessMaterial, 0, True)
        draw_rectangle(rIn2, length, thicknessMaterial, cutAngle)

    if theta_last > 0:
        currentAngle = math.floor(nTurns) * 2 * math.pi
        rIn = a + b * currentAngle
        currentAngle2 = math.floor(nTurns) * 2 * math.pi + min(cutAngle, theta_last)
        rIn2 = a + b * currentAngle2

        angleToGo = min(cutAngle, theta_last)
        draw_spiralSheet(rIn, thicknessSpiral, length, angleToGo, n, True)
        draw_spiralSheet(rIn + thicknessMaterial, thicknessSpiral, length, angleToGo, n, False)
        draw_spiralFront(rIn, thicknessMaterial, thicknessSpiral, 0, angleToGo, n, True)
        draw_spiralFront(rIn, thicknessMaterial, thicknessSpiral, length, angleToGo, n)
        draw_rectangle(rIn, length, thicknessMaterial, 0, True)
        draw_rectangle(rIn2, length, thicknessMaterial, angleToGo)

    glPopMatrix()


def draw_simple_rectangle(width, height):
    glRectf(0, 0, width, height)


def draw_rectangle(rIn, length, thickness, angle, reverseDirection=False):
    if reverseDirection:
        glFrontFace(GL_CW)

    glPushMatrix()
    if reverseDirection:
        glRotatef(-90, 0, 1, 0)
        glRotatef(-90, 1, 0, 0)
    else:
        glRotatef(-90, 0, 1, 0)
        glRotatef(-90, 1, 0, 0)
    glRotatef(angle * 180 / math.pi, 1, 0, 0)
    glRectf(length, thickness + rIn, 0, rIn)
    glPopMatrix()

    glFrontFace(GL_CCW)


def draw_2Dring(innerRadius, outerRadius, z0, theAngle, n, reverseDirection=False):
    if 2 * math.pi * 0.95 < theAngle < 2 * math.pi * 1.05:
        draw_disk(innerRadius, outerRadius, n, z0)
    else:
        glBegin(GL_QUAD_STRIP)
        numberOfPoints = n
        rIn = innerRadius
        rOut = outerRadius
        for i in range(numberOfPoints + 1):
            if reverseDirection:
                angle = (n - i) / numberOfPoints * theAngle
            else:
                angle = i / numberOfPoints * theAngle
            glVertex3f(rIn * math.cos(angle), rIn * math.sin(angle), z0)
            glVertex3f(rOut * math.cos(angle), rOut * math.sin(angle), z0)
            glNormal3f(0, 0, 1)
        glEnd()


def draw_2Dring_diff_angle(innerRadius, outerRadius, angle_in, angle_out, n, reverseDirection=False):
    glBegin(GL_QUAD_STRIP)
    numberOfPoints = n
    rIn = innerRadius
    rOut = outerRadius
    for i in range(numberOfPoints + 1):
        if reverseDirection:
            angle1 = -angle_in/2 + (n - i) / numberOfPoints * angle_in
            angle2 = -angle_out/2 + (n - i) / numberOfPoints * angle_out
        else:
            angle1 = -angle_in/2 + i / numberOfPoints * angle_in
            angle2 = -angle_out/2 + i / numberOfPoints * angle_out
        glVertex3f(rIn * math.cos(angle1), rIn * math.sin(angle1), 0)
        glVertex3f(rOut * math.cos(angle2), rOut * math.sin(angle2), 0)
        glNormal3f(0, 0, 1)
    glEnd()


def draw_tubeSheet(radius, length, theAngle, n, reverseDirection=False):
    circle_pts = []
    for i in range(n + 1):
        if reverseDirection:
            angle = -theAngle/2 + theAngle * ((n - i) / n)
        else:
            angle = -theAngle/2 + theAngle * (i / n)
        x = radius * math.cos(angle)
        y = radius * math.sin(angle)
        pt = [x, y]
        circle_pts.append(pt)

    glBegin(GL_TRIANGLE_STRIP)
    for [x, y] in circle_pts:
        if reverseDirection:
            glVertex(x, y, length)
            glVertex(x, y, 0)

        else:
            glVertex(x, y, length)
            glVertex(x, y, 0)
        glNormal3f(x, y, 0)
    glEnd()


def draw_cylinder(innerRadius, outerRadius, length, n, translate=0):
    glPushMatrix()
    glTranslate(0, 0, translate)
    rIn = innerRadius
    rOut = outerRadius
    gluCylinder(gluNewQuadric(), rIn, rIn, length, n, n)  # Lightning ko but interface ok :x
    gluCylinder(gluNewQuadric(), rOut, rOut, length, n, n)
    gluDisk(gluNewQuadric(), rIn, rOut, n, n)
    glTranslate(0, 0, length)
    gluDisk(gluNewQuadric(), rIn, rOut, n, n)
    glPopMatrix()


# Angle in radians
def draw_part_cylinder(innerRadius, outerRadius, length, angle, n, translate=0, drawSides=True):
    if 2 * math.pi * 0.95 < angle < 2 * math.pi * 1.05:
        draw_cylinder(innerRadius, outerRadius, length, n, translate)
    else:
        glPushMatrix()
        glTranslate(0, 0, translate - length / 2)

        rIn = innerRadius
        rOut = outerRadius
        if drawSides:
            draw_tubeSheet(rIn, length, angle, n, True)
            draw_tubeSheet(rOut, length, angle, n, False)
        draw_2Dring(rIn, rOut, 0, angle, n, True)
        draw_2Dring(rIn, rOut, length, angle, n, False)
        glRotatef(-90, 0, 1, 0)
        glRotatef(-90, 1, 0, 0)

        glFrontFace(GL_CW)
        glRectf(length, outerRadius, 0, 0 + innerRadius)
        glFrontFace(GL_CCW)

        glRotatef(angle * 180 / math.pi, 1, 0, 0)
        glRectf(length, outerRadius, 0, 0 + innerRadius)
        glPopMatrix()


# Angle in radians
def draw_disk(innerRadius, outerRadius, n, translate=0):
    glPushMatrix()
    glTranslate(0, 0, translate)
    gluDisk(gluNewQuadric(), innerRadius, outerRadius, n, n)
    glPopMatrix()


# Angle in radians
def draw_part_disk(innerRadius, outerRadius, thickness, angle, n, translate=0):
    glPushMatrix()
    glTranslate(0, 0, translate)

    rIn = innerRadius
    rOut = outerRadius
    draw_tubeSheet(rIn, thickness, angle, n, True)
    draw_tubeSheet(rOut, thickness, angle, n, False)
    draw_2Dring(rIn, rOut, 0, angle, n, True)
    draw_2Dring(rIn, rOut, thickness, angle, n, False)
    glRotatef(-90, 0, 1, 0)
    glRotatef(-90, 1, 0, 0)

    glFrontFace(GL_CW)
    glRectf(thickness, rOut, 0, 0 + innerRadius)
    glFrontFace(GL_CCW)

    glRotatef(angle * 180 / math.pi, 1, 0, 0)
    glRectf(thickness, rOut, 0, 0 + innerRadius)
    glPopMatrix()


# Angle in radians
def draw_part_disk_diff_angles(innerRadius, outerRadius, thickness, angle_in, angle_out, n):
    rIn = innerRadius
    rOut = outerRadius
    draw_tubeSheet(rIn, thickness, angle_in, n, True)
    draw_tubeSheet(rOut, thickness, angle_out, n, False)

    draw_2Dring_diff_angle(innerRadius, outerRadius, angle_in, angle_out, n, True)

    glPushMatrix()
    glTranslate(0, 0, thickness)
    draw_2Dring_diff_angle(innerRadius, outerRadius, angle_in, angle_out, n, False)
    glPopMatrix()

    for dir_num, dir_angle in [(GL_CW, 1), (GL_CCW, -1)]:
        glFrontFace(dir_num)
        glBegin(GL_QUAD_STRIP)
        glVertex3f(rIn * math.cos(dir_angle*angle_in/2), rIn * math.sin(dir_angle*angle_in/2), 0)
        glVertex3f(rIn * math.cos(dir_angle*angle_in/2), rIn * math.sin(dir_angle*angle_in/2), thickness)
        glVertex3f(rOut * math.cos(dir_angle*angle_out/2), rOut * math.sin(dir_angle*angle_out/2), 0)
        glVertex3f(rOut * math.cos(dir_angle*angle_out/2), rOut * math.sin(dir_angle*angle_out/2), thickness)
        glEnd()


# Angle in radians
def draw_carved_disk(innerRadius, outerRadius, carvedRin, carvedRout, thickness, depth, angle, n, translate=0):
    glPushMatrix()
    glTranslate(0, 0, translate)

    rIn = innerRadius
    rOut = outerRadius
    rInOut = carvedRin
    rOutIn = carvedRout

    if depth > thickness:
        if rInOut > rIn:
            draw_tubeSheet(rIn, thickness, angle, n, True)
            draw_tubeSheet(rInOut, thickness, angle, n, False)
            draw_2Dring(rIn, rInOut, 0, angle, n, True)
            draw_2Dring(rIn, rInOut, thickness, angle, n, False)
        if rOut > rOutIn:
            draw_tubeSheet(rOutIn, thickness, angle, n, True)
            draw_tubeSheet(rOut, thickness, angle, n, False)
            draw_2Dring(rOutIn, rOut, 0, angle, n, True)
            draw_2Dring(rOutIn, rOut, thickness, angle, n, False)

        glRotatef(-90, 0, 1, 0)
        glRotatef(-90, 1, 0, 0)

        glFrontFace(GL_CW)
        if rOut > rOutIn:
            glRectf(thickness, rOut, 0, rOutIn)
        if rInOut > rIn:
            glRectf(thickness, rInOut, 0, rIn)
        glFrontFace(GL_CCW)

        glRotatef(angle * 180 / math.pi, 1, 0, 0)
        if rOut > rOutIn:
            glRectf(thickness, rOut, 0, rOutIn)
        if rInOut > rIn:
            glRectf(thickness, rInOut, 0, rIn)
        glPopMatrix()

    else:
        extremeIn = rInOut
        extremeOut = rOutIn

        if rInOut <= rIn:
            extremeIn = rIn
            draw_tubeSheet(rIn, thickness - depth, angle, n, True)
        else:
            draw_tubeSheet(rIn, thickness, angle, n, True)
            draw_tubeSheet(rInOut, thickness, angle, n, False)
            draw_2Dring(rIn, rInOut, thickness, angle, n, False)

        if rOutIn >= rOut:
            extremeIn = rOut
            draw_tubeSheet(rOut, thickness - depth, angle, n, False)
        else:
            draw_tubeSheet(rOut, thickness, angle, n, False)
            draw_tubeSheet(rOutIn, thickness, angle, n, True)
            draw_2Dring(rOutIn, rOut, thickness, angle, n, False)

        draw_2Dring(extremeIn, extremeOut, thickness - depth, angle, n, False)
        draw_2Dring(rIn, rOut, 0, angle, n, True)

        glRotatef(-90, 0, 1, 0)
        glRotatef(-90, 1, 0, 0)

        glFrontFace(GL_CW)
        glRectf(thickness - depth, extremeOut, 0, extremeIn)
        if rOut > extremeOut:
            glRectf(thickness, rOut, 0, extremeOut)
        if rIn < extremeIn:
            glRectf(thickness, extremeIn, 0, rIn)
        glFrontFace(GL_CCW)

        glRotatef(angle * 180 / math.pi, 1, 0, 0)
        glRectf(thickness - depth, extremeOut, 0, extremeIn)
        if rOut > extremeOut:
            glRectf(thickness, rOut, 0, extremeOut)
        if rIn < extremeIn:
            glRectf(thickness, extremeIn, 0, rIn)
        glPopMatrix()


def draw_part_cylinder_throat(rIn, rOut, rOutThroat, length, lengthThroat, angle, n, translate=0):
    glPushMatrix()
    glTranslate(0, 0, translate - length / 2)

    glPushMatrix()
    glTranslate(0, 0, (length + lengthThroat) / 2)
    draw_tubeSheet(rOut, (length - lengthThroat) / 2, angle, n, False)
    glPopMatrix()

    glPushMatrix()
    glTranslate(0, 0, (length - lengthThroat) / 2)
    draw_tubeSheet(rOutThroat, lengthThroat, angle, n, False)
    glPopMatrix()

    draw_tubeSheet(rOut, (length - lengthThroat) / 2, angle, n, False)

    draw_tubeSheet(rIn, length, angle, n, True)

    draw_2Dring(rIn, rOut, length, angle, n, False)  # weird shadow :(
    draw_2Dring(rOutThroat, rOut, (length + lengthThroat) / 2, angle, n, True)
    draw_2Dring(rOutThroat, rOut, (length - lengthThroat) / 2, angle, n, False)
    draw_2Dring(rIn, rOut, 0, angle, n, True)

    draw_rectangle(rIn, (length - lengthThroat) / 2, rOut - rIn, 0, True)

    glPushMatrix()
    glTranslate(0, 0, (length - lengthThroat) / 2)
    draw_rectangle(rIn, lengthThroat, rOutThroat - rIn, 0, True)
    glPopMatrix()

    glPushMatrix()
    glTranslate(0, 0, (length + lengthThroat) / 2)
    draw_rectangle(rIn, (length - lengthThroat) / 2, rOut - rIn, 0, True)
    glPopMatrix()

    draw_rectangle(rIn, (length - lengthThroat) / 2, rOut - rIn, angle, False)

    glPushMatrix()
    glTranslate(0, 0, (length - lengthThroat) / 2)
    draw_rectangle(rIn, lengthThroat, rOutThroat - rIn, angle, False)
    glPopMatrix()

    glPushMatrix()
    glTranslate(0, 0, (length + lengthThroat) / 2)
    draw_rectangle(rIn, (length - lengthThroat) / 2, rOut - rIn, angle, False)
    glPopMatrix()

    glPopMatrix()


def drawWireTube(diameter, xa, ya, xb, yb, n=50, translateZ=0):
    glPushMatrix()
    glTranslate(0, 0, translateZ)
    glTranslate(xa, ya, 0)

    vec = xb - xa, yb - ya
    Norm = math.sqrt(vec[0] ** 2 + vec[1] ** 2)
    vecN = vec[0] / Norm, vec[1] / Norm
    phi = math.acos(math.fabs(vecN[1]))

    if xb < xa:
        if yb > ya:
            Phi = phi
        else:
            Phi = math.pi - phi
    else:
        if yb > ya:
            Phi = - phi
        else:
            Phi = -math.pi + phi
    glRotatef(Phi * 180 / math.pi, 0, 0, 1)
    glRotatef(-90, 1, 0, 0)
    gluCylinder(gluNewQuadric(), diameter / 2, diameter / 2, Norm, n, n)
    glPopMatrix()



from .onselectInterface import OnselectInterface
from optimeed.core import ListDataStruct


class Onselect_splitTrace(OnselectInterface):
    def __init__(self, theLinkDataGraphs):
        """Create two new traces from selections (selected and not selected)"""
        self.theLinkDataGraphs = theLinkDataGraphs
        self.selectedTraces = dict()

    def selector_updated(self, selection_name, the_collection, selected_data, not_selected_data):
        """
        Action to perform once the data have been selected

        :param selection_name: name of the selection (deprecated ?)
        :param the_collection: the collection
        :param selected_data: indices of the data selected
        :param not_selected_data: indices of the data not selected
        :return: identifier that can later be used with cancel_selector
        """
        selections = list()
        for data, name in [(selected_data, "Selected"), (not_selected_data, "Not selected")]:
            newStruct = ListDataStruct()
            theData = [the_collection.get_data_at_index(index) for index in data]
            newStruct.set_data(theData)
            selections.append(self.theLinkDataGraphs.add_collection(newStruct, {"legend": "{} | {}".format(selection_name, name)}))

        self.theLinkDataGraphs.update_graphs()

        return selections

    def cancel_selector(self, selection_identifiers):
        for selection_identifier in selection_identifiers:
            self.theLinkDataGraphs.remove_collection(selection_identifier)

    def get_name(self):
        return "Split trace"



from .onselectInterface import OnselectInterface
from .onselect_highlight import Onselect_highlight
from .onselect_newTrace import Onselect_newTrace
from .onselect_splitTrace import Onselect_splitTrace



from optimeed.visualize.graphs.widget_graphsVisual import Widget_graphsVisual
from optimeed.visualize.mainWindow import MainWindow
from optimeed.visualize.process_mainloop import start_qt_mainloop
from optimeed.core.graphs import Data, Graphs
from optimeed.visualize.onclick import Onclick_measure


class _PlotHolders:
    def __init__(self):
        self.curr_g = None
        self.graphs = Graphs()
        self.theGraphVisual = Widget_graphsVisual(self.graphs, refresh_time=-1, is_light=True)
        self.list_actions_on_clik = list()
        # self.new_plot()

    def add_plot(self, x, y, **kwargs):
        if self.curr_g is None:
            self.new_plot()
        theData = Data(x, y, **kwargs)
        return self.curr_g, self.graphs.add_trace(self.curr_g, theData)

    def get_wgGraphs(self):
        return self.theGraphVisual

    def new_plot(self):
        self.curr_g = self.graphs.add_graph()

    def set_title(self, theTitle, **kwargs):
        self.theGraphVisual.set_title(self.curr_g, theTitle, **kwargs)

    def reset(self):
        self.graphs.reset()
        self.curr_g = None

    def axis_equal(self):
        self.theGraphVisual.get_graph(self.curr_g).axis_equal()


class WindowHolders:
    def __init__(self):
        self.figures = dict()
        self.plotHolders = dict()
        self.currFigure = None
        self.actions_on_click = [Onclick_measure()]

    def set_currFigure(self, currFigure):
        if currFigure is None:
            self.new_figure()
        else:
            self.currFigure = currFigure

    def add_plot(self, *args, **kwargs):
        if self.currFigure is None:
            self.currFigure = 0
            self.new_figure()
        graph_id, trace_id = self.get_curr_plotHolder().add_plot(*args, **kwargs)
        return self.currFigure, graph_id, trace_id

    def set_title(self, *args, **kwargs):
        self.get_curr_plotHolder().set_title(*args, **kwargs)

    def new_figure(self):
        self.currFigure += 1
        self.plotHolders[self.currFigure] = _PlotHolders()

        def actionOnWindowClose(mainWindow, _):
            # stop_qt_mainloop()
            mainWindow.close()
            self.plotHolders[self.currFigure].reset()
            self.currFigure = None
            # self.figures[self.currFigure] = self.create_figure()

        # guiPyqtgraph(self.plotHolders[self.currFigure].get_wgGraphs(), actionsOnClick=self.actions_on_click)
        the_mainWindow = MainWindow([self.plotHolders[self.currFigure].get_wgGraphs()], size=[1000, 700], actionOnWindowClosed=actionOnWindowClose)
        self.figures[self.currFigure] = the_mainWindow
        the_mainWindow.run(False)

    def new_plot(self):
        self.get_curr_plotHolder().new_plot()

    def show(self):
        self.figures[self.currFigure].hold()

    def get_curr_plotHolder(self):
        return self.plotHolders[self.currFigure]

    def get_wgGraphs(self, fig=None):
        if fig is None:
            return self.get_curr_plotHolder().get_wgGraphs()
        return self.plotHolders[fig].get_wgGraphs()

    def get_all_figures(self):
        return list(self.figures.keys())

    def axis_equal(self):
        self.get_curr_plotHolder().axis_equal()

    def add_action_on_click(self, theAction):
        self.actions_on_click.append(theAction)


myWindows = WindowHolders()


def plot(x, y, hold=False, **kwargs):
    """Plot new trace"""
    idPlot = myWindows.add_plot(x, y, **kwargs)

    if hold:
        show()

    return idPlot


def show():
    """Show (start qt mainloop) graphs. Blocking"""
    start_qt_mainloop()
    # myWindows.show()


def figure(numb=None):
    """Set current figure"""
    myWindows.set_currFigure(numb)

def add_action_on_click(theAction):
    myWindows.add_action_on_click(theAction)

def set_title(theTitle, **kwargs):
    """Set title of the plot"""
    myWindows.set_title(theTitle, **kwargs)


def axis_equal():
    myWindows.axis_equal()


def get_all_figures():
    """Get all existing figures"""
    return myWindows.get_all_figures()


def get_wgGraphs(fig=None):
    """Advanced option.
    :return: :class:`~optimeed.visualize.gui.widgets.widget_graphs_visual.widget_graphs_visual`
    """
    return myWindows.get_wgGraphs(fig)




from .parametric_analysis import *
from .fit import leastSquare, do_fit
from .sensitivity_analysis import get_sensitivity_problem, evaluate_sensitivities, SensitivityParameters, analyse_sobol_convergence
from .sensitivity_analysis import prepare_embarrassingly_parallel_sensitivity, gather_embarrassingly_parallel_sensitivity, launch_embarrassingly_parallel_sensitivity



# Manage inkscape options, used in exporting graphs
import subprocess
import re
import platform
from .tools import printIfShown, SHOW_WARNING
from subprocess import DEVNULL, STDOUT


def get_path_to_inkscape():
    platform_name = platform.system()
    if platform_name == "Windows":
        return '"C:\Program Files\Inkscape\\bin\inkscape.com"'
    return "inkscape"


def get_inkscape_version():
    converter_cmd_line = f'{get_path_to_inkscape()} --version'
    result = subprocess.run(converter_cmd_line, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    result = str(result.stdout)
    res = re.search(r'([\d].[\d])', result)  # Match the version number
    try:
        return float(res.group(0))
    except AttributeError:
        return None


inkscape_version = get_inkscape_version()


def inkscape_svg_to_pdf(filename_svg, filename_pdf):
    if inkscape_version is None:
        printIfShown("Inkscape not available for pdf export", SHOW_WARNING)
        return
    elif inkscape_version < 1.0:
        converter_cmd_line = get_path_to_inkscape() + ' -z {} -A {} --export-area-drawing'.format(filename_svg, filename_pdf)
    else:
        converter_cmd_line = get_path_to_inkscape() + ' --export-type=pdf {} -o {} --export-area-drawing'.format(filename_svg, filename_pdf)
    subprocess.Popen(converter_cmd_line, shell=True, stdout=DEVNULL, stderr=STDOUT)


def inkscape_svg_to_png(filename_svg, filename_png):
    if inkscape_version is None:
        printIfShown("Inkscape not available for png export", SHOW_WARNING)
        return
    elif inkscape_version < 1.0:
        converter_cmd_line = get_path_to_inkscape() + ' -z {} -D -e {} -d 400 --export-area-drawing'.format(filename_svg, filename_png)
    else:
        converter_cmd_line = get_path_to_inkscape() + ' --export-type=png {} -o {} -d 400 --export-area-drawing'.format(filename_svg, filename_png)
    subprocess.Popen(converter_cmd_line, shell=True, stdout=DEVNULL, stderr=STDOUT)


from PyQt5 import QtWidgets, QtGui, QtCore
import base64


class Widget_image(QtWidgets.QLabel):

    def __init__(self, image_b64):
        """This widget allows to easily visualize any png image and manages redimensioning

        :param image_b64: An encoded base 64 png, typically base64(file.read())
        """

        QtWidgets.QLabel.__init__(self)
        self.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Ignored)
        self.initialPixmap = QtGui.QPixmap()
        self.set_image(image_b64)
        self.installEventFilter(self)
        self.setAlignment(QtCore.Qt.AlignCenter)

    def eventFilter(self, source, event):
        if source is self and event.type() == QtCore.QEvent.Resize:
            self.setPixmap(self.initialPixmap.scaled(self.size()*0.95, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation))
        return super().eventFilter(source, event)

    def set_image(self, image_b64):
        """ Set new image to widget """

        self.initialPixmap.loadFromData(base64.b64decode(image_b64))
        self.setPixmap(self.initialPixmap)

